import asyncio
import random
from typing import Any, Dict, Optional

from unsserv.common.services_abc import ISearchingService, IMembershipService
from unsserv.common.structs import Node, Property
from unsserv.common.utils import get_random_id
from unsserv.extreme.searching.config import KWalkerConfig
from unsserv.extreme.searching.protocol import KWalkerProtocol
from unsserv.extreme.searching.structs import Walk, WalkResult


class KWalker(ISearchingService):
    properties = {Property.EXTREME}
    _protocol: KWalkerProtocol
    _config: KWalkerConfig

    _search_data: Dict[str, bytes]
    _walk_events: Dict[str, asyncio.Event]
    _walk_results: Dict[str, bytes]

    def __init__(self, membership: IMembershipService):
        self.membership = membership
        self.my_node = membership.my_node
        self._protocol = KWalkerProtocol(self.my_node)
        self._config = KWalkerConfig()

        self._search_data = {}
        self._walk_results = {}
        self._walk_events = {}

    async def join(self, service_id: str, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Searching service")
        self.service_id = service_id
        await self._initialize_protocol()
        self._config.load_from_dict(configuration)
        self._search_data = {}
        self.running = True

    async def leave(self):
        if not self.running:
            return
        await self._protocol.stop()
        self.running = False

    async def publish(self, data_id: str, data: bytes):
        if not self.running:
            raise RuntimeError("Searching service not running")
        if data_id in self._search_data:
            raise KeyError("Data-id already published")
        if not isinstance(data, bytes):
            raise TypeError("Published data must be bytes type")
        if not isinstance(data_id, str):
            raise TypeError("Published data-id must be str type")

        self._search_data[data_id] = data

    async def unpublish(self, data_id: str):
        if not self.running:
            raise RuntimeError("Searching service not running")
        if data_id not in self._search_data:
            raise KeyError("Data-id not published")
        del self._search_data[data_id]

    async def search(self, data_id: str) -> Optional[bytes]:
        candidate_neighbours = self.membership.get_neighbours()
        assert isinstance(candidate_neighbours, list)
        fanout = min(self._config.FANOUT, len(candidate_neighbours))
        walk_id = get_random_id()
        walk = Walk(
            id=walk_id, data_id=data_id, origin_node=self.my_node, ttl=self._config.TTL
        )
        self._walk_events[walk_id] = asyncio.Event()
        for neighbour in random.sample(candidate_neighbours, fanout):
            await self._protocol.walk(neighbour, walk)
        try:
            return await asyncio.wait_for(
                self._get_walk_result(fanout, walk_id), timeout=self._config.TIMEOUT
            )
        except asyncio.TimeoutError:
            return None

    async def _get_walk_result(self, fanout, walk_id):
        results_amount = 0
        result = None
        while results_amount < fanout:
            await self._walk_events[walk_id].wait()
            results_amount += 1
            result = self._walk_results[walk_id]
            if result:
                return result
        return result

    async def _handler_walk(self, sender: Node, walk: Walk):
        result = self._search_data.get(walk.data_id, None)
        if result or walk.ttl < 1:
            walk_result = WalkResult(walk_id=walk.id, result=result)
            asyncio.create_task(
                self._protocol.walk_result(walk.origin_node, walk_result)
            )
        else:
            next_walk = Walk(
                id=walk.id,
                data_id=walk.data_id,
                origin_node=walk.origin_node,
                ttl=walk.ttl - 1,
            )
            candidate_neighbours = self.membership.get_neighbours()
            assert isinstance(candidate_neighbours, list)
            neighbour = random.choice(candidate_neighbours)
            asyncio.create_task(self._protocol.walk(neighbour, next_walk))

    async def _handler_walk_result(self, sender: Node, walk_result: WalkResult):
        self._walk_results[walk_result.walk_id] = walk_result.result
        self._walk_events[walk_result.walk_id].set()

    async def _initialize_protocol(self):
        self._protocol.set_handler_walk(self._handler_walk)
        self._protocol.set_handler_walk_result(self._handler_walk_result)
        await self._protocol.start(self.service_id)



import asyncio
from collections import Counter
from math import ceil

import pytest

from tests.utils import get_random_nodes
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.stable.membership.hyparview import HyParView

MEMBERSHIP_SERVICE_ID = "hyparview"
node = Node(("127.0.0.1", 7771))


@pytest.mark.asyncio
@pytest.fixture
async def init_hyparview():
    hyparview = None
    r_hyparviews = []

    async def _init_hyparview(amount):
        nonlocal hyparview, r_hyparviews
        hyparview = HyParView(node)
        await hyparview.join(MEMBERSHIP_SERVICE_ID)
        r_nodes = get_random_nodes(amount)
        for i, r_node in enumerate(r_nodes):
            r_hyparview = HyParView(r_node)
            await r_hyparview.join(
                MEMBERSHIP_SERVICE_ID, bootstrap_nodes=[node] + r_nodes[:i]
            )
            r_hyparviews.append(r_hyparview)
        return hyparview, r_hyparviews

    try:
        yield _init_hyparview
    finally:
        await hyparview.leave()
        for r_hyparview in r_hyparviews:
            await r_hyparview.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_join_hyparview(init_hyparview, amount):
    hyparview, r_hyparviews = await init_hyparview(amount)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)

    all_nodes = set(
        [
            item
            for sublist in map(lambda n: n.get_neighbours(), r_hyparviews + [hyparview])
            for item in sublist
        ]
    )
    assert amount < len(all_nodes)

    all_neighbours = {}
    for r_hyparview in r_hyparviews + [hyparview]:
        all_neighbours[r_hyparview.my_node] = r_hyparview.get_neighbours()
    for node, neighbours in all_neighbours.items():
        for neighbour in neighbours:
            assert node in all_neighbours[neighbour]  # check symmetry


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        GossipConfig.LOCAL_VIEW_SIZE + 1,
        GossipConfig.LOCAL_VIEW_SIZE + 5,
        GossipConfig.LOCAL_VIEW_SIZE + 100,
    ],
)
async def test_leave_hyparview(init_hyparview, amount):
    hyparview, r_hyparviews = await init_hyparview(amount)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)

    await hyparview.leave()
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 40)

    all_nodes = Counter(
        [
            item
            for sublist in map(lambda n: n.get_neighbours(), r_hyparviews)
            for item in sublist
        ]
    )
    nodes_ten_percent = ceil(amount * 0.2)
    assert hyparview.my_node not in all_nodes.keys() or hyparview.my_node in set(
        map(lambda p: p[0], all_nodes.most_common()[-nodes_ten_percent:])
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 1,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 5,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 100,
    ],
)  # very high neighbours amount,
# to assure neighbours will change, because it is initailzied by Newscast
async def test_hyparview_handler(init_hyparview, amount):
    hyparview, r_hyparviews = await init_hyparview(amount)

    handler_event = asyncio.Event()

    async def handler(local_view):
        assert isinstance(local_view, list)
        nonlocal handler_event
        handler_event.set()

    hyparview.add_neighbours_handler(handler)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)
    assert handler_event.is_set()



import asyncio
import math
import random
from abc import ABC, abstractmethod
from collections import Counter
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Set, Tuple

from unsserv.common.gossip.config import (
    GossipConfig,
    ViewSelectionPolicy,
    PeerSelectionPolicy,
    ViewPropagationPolicy,
)
from unsserv.common.gossip.protocol import GossipProtocol
from unsserv.common.gossip.structs import PushData
from unsserv.common.gossip.typing import ExternalViewSource, CustomSelectionRanking
from unsserv.common.gossip.typing import Payload, View
from unsserv.common.structs import Node
from unsserv.common.typing import Handler
from unsserv.common.utils import stop_task, HandlersManager


class IGossipSubscriber(ABC):
    @abstractmethod
    async def receive_payload(self, payload: Payload):
        pass

    @abstractmethod
    async def get_payload(self) -> Tuple[Any, Any]:
        pass


class Gossip:
    my_node: Node
    local_view: View
    running: bool = False
    _config: GossipConfig
    _protocol: GossipProtocol
    _handler_manager: HandlersManager

    def __init__(
        self,
        my_node: Node,
        service_id: Any,
        local_view_nodes: List[Node] = None,
        local_view_handler: Handler = None,
        external_nodes_source: ExternalViewSource = None,
        custom_selection_ranking: CustomSelectionRanking = None,
        **configuration
    ):
        self.my_node = my_node
        self.service_id = service_id
        self._config = GossipConfig()
        self._config.load_from_dict(configuration)
        self._protocol = GossipProtocol(self.my_node)

        self.local_view = Counter(
            random.sample(
                local_view_nodes,
                min(len(local_view_nodes), self._config.LOCAL_VIEW_SIZE),
            )
            if local_view_nodes
            else []
        )

        self.get_external_nodes = external_nodes_source
        self.custom_selection_ranking = custom_selection_ranking

        self.subscribers: List[IGossipSubscriber] = []
        self._handler_manager = HandlersManager()
        if local_view_handler:
            self._handler_manager.add_handler(local_view_handler)

    async def start(self):
        if self.running:
            raise RuntimeError("Already running Gossip")
        await self._initialize_protocol()
        self._gossip_task = asyncio.create_task(self._gossip_loop())
        self.running = True

    async def stop(self):
        if not self.running:
            return
        await self._protocol.stop()
        if self._gossip_task:
            await stop_task(self._gossip_task)
        self.running = False

    def subscribe(self, subscriber: IGossipSubscriber):
        self.subscribers.append(subscriber)

    def unsubscribe(self, subscriber: IGossipSubscriber):
        self.subscribers.remove(subscriber)

    async def _gossip_loop(self):
        while True:
            old_neighbours = set(self.local_view.keys())
            await asyncio.sleep(self._config.GOSSIPING_FREQUENCY)
            await self._exchange_with_peer()
            self._call_handler_if_view_changed(old_neighbours)

    async def _exchange_with_peer(self):
        peer = self._select_peer(self.local_view)
        if not peer:  # empty Local view
            return
        subscribers_data = await self._get_data_from_subscribers()
        if self._config.VIEW_PROPAGATION is ViewPropagationPolicy.PUSH:
            push_view = self._merge_views(Counter({self.my_node: 0}), self.local_view)
            push_data = PushData(view=push_view, payload=subscribers_data)
            with self._remove_on_connection_error(peer):
                await self._protocol.push(peer, push_data)
        elif self._config.VIEW_PROPAGATION is ViewPropagationPolicy.PULL:
            with self._remove_on_connection_error(peer):
                pull_response = await self._protocol.pull(peer, subscribers_data)
                await self._handler_push(peer, pull_response)
        elif self._config.VIEW_PROPAGATION is ViewPropagationPolicy.PUSHPULL:
            push_view = self._merge_views(Counter({self.my_node: 0}), self.local_view)
            push_data = PushData(view=push_view, payload=subscribers_data)
            with self._remove_on_connection_error(peer):
                pull_response = await self._protocol.pushpull(peer, push_data)
                await self._handler_push(peer, pull_response)

    def _select_peer(self, view: View) -> Optional[Node]:
        if self.custom_selection_ranking:
            ordered_nodes = self.custom_selection_ranking(view)
        else:
            ordered_nodes = list(map(lambda n: n[0], reversed(view.most_common())))
        if self._config.PEER_SELECTION is PeerSelectionPolicy.RAND:
            return random.choice(ordered_nodes) if view else None
        elif self._config.PEER_SELECTION is PeerSelectionPolicy.HEAD:
            return ordered_nodes[0] if view else None
        elif self._config.PEER_SELECTION is PeerSelectionPolicy.TAIL:
            return ordered_nodes[-1] if view else None
        raise AttributeError("Invalid Peer Selection policy")

    def _select_view(self, view: View) -> View:
        # todo: document what happens here
        if self.my_node in view:
            view.pop(self.my_node)
        remove_amount = max(len(view) - self._config.LOCAL_VIEW_SIZE, 0)
        if remove_amount < 1:
            return view
        if self.custom_selection_ranking:
            ordered_nodes = self.custom_selection_ranking(view)
        else:
            ordered_nodes = list(map(lambda n: n[0], reversed(view.most_common())))

        new_view = view.copy()
        if self._config.VIEW_SELECTION is ViewSelectionPolicy.RAND:
            for node in random.sample(ordered_nodes, remove_amount):
                new_view.pop(node)
        elif self._config.VIEW_SELECTION is ViewSelectionPolicy.HEAD:
            for node in ordered_nodes[-remove_amount:]:
                new_view.pop(node)
        elif self._config.VIEW_SELECTION is ViewSelectionPolicy.TAIL:
            for node in ordered_nodes[:remove_amount]:
                new_view.pop(node)
        else:
            raise AttributeError("Invalid View Selection policy")
        return new_view

    def _increase_hop_count(self, view: View) -> View:
        return view + Counter(view.keys())

    def _merge_views(self, view1: View, view2: View) -> View:
        all_nodes = set(list(view1.keys()) + list(set(view2.keys())))
        merged_view: Counter = Counter()
        for node in all_nodes:
            merged_view[node] = (
                view1[node]
                if view1.get(node, math.inf) < view2.get(node, math.inf)
                else view2[node]
            )
        return merged_view

    async def _get_data_from_subscribers(self):
        data: Dict = {}
        for subscriber in self.subscribers:
            key, value = await subscriber.get_payload()
            data[key] = value
        return data

    async def _deliver_message_to_subscribers(self, gossip_payload: Payload):
        for subscriber in self.subscribers:
            await subscriber.receive_payload(gossip_payload)

    def _call_handler_if_view_changed(self, old_neighbours: Set):
        current_neighbours = set(self.local_view.keys())
        if old_neighbours == current_neighbours:
            return
        self._handler_manager.call_handlers(self.local_view)

    async def _handler_push(self, sender: Node, push_data: PushData):
        view = self._increase_hop_count(push_data.view)
        buffer = self._merge_views(view, self.local_view)
        if self.get_external_nodes:
            buffer = self._merge_views(view, Counter(self.get_external_nodes()))
        new_view = self._select_view(buffer)
        self.local_view = new_view
        await self._deliver_message_to_subscribers(push_data.payload)

    async def _handler_pull(self, sender: Node, payload: Payload) -> PushData:
        my_descriptor = Counter({self.my_node: 0})
        subscribers_data = await self._get_data_from_subscribers()
        # todo: deliver message to subscribers
        view = self._merge_views(my_descriptor, self.local_view)
        return PushData(view=view, payload=subscribers_data)

    async def _handler_pushpull(self, sender: Node, push_data: PushData):
        pull_response = await self._handler_pull(sender, push_data.payload)
        await self._handler_push(sender, push_data)
        return pull_response

    async def _initialize_protocol(self):
        self._protocol.set_handler_push(self._handler_push)
        self._protocol.set_handler_pull(self._handler_pull)
        self._protocol.set_handler_pushpull(self._handler_pushpull)
        await self._protocol.start(self.service_id)

    @contextmanager
    def _remove_on_connection_error(self, node: Node):
        try:
            yield
        except ConnectionError:
            if node in self.local_view:
                self.local_view.pop(node)



from enum import IntEnum, auto
from typing import Tuple, Sequence, Any, List

from unsserv.common.utils import parse_node
from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.extreme.dissemination.lpbcast.structs import Event
from unsserv.extreme.dissemination.lpbcast.typing import Digest

FIELD_COMMAND = "lpbcast-command"
FIELD_EVENT_DATA = "lpbcast-event-data"
FIELD_EVENT_ID = "lpbcast-event-id"
FIELD_EVENT_ORIGIN = "lpbcast-event-origin"
FIELD_DIGEST = "lpbcast-digest"
FIELD_RETRIEVE_EVENT = "lpbcast-retrieve-events"


class LpbcastCommand(IntEnum):
    PUSH_EVENT = auto()
    RETRIEVE_EVENT = auto()


class LpbcastTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == LpbcastCommand.PUSH_EVENT:
            event: Event = data[0]
            message_data = {
                FIELD_COMMAND: LpbcastCommand.PUSH_EVENT,
                FIELD_EVENT_ID: event.id,
                FIELD_EVENT_DATA: event.data,
                FIELD_EVENT_ORIGIN: event.origin,
                FIELD_DIGEST: event.digest,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == LpbcastCommand.RETRIEVE_EVENT:
            event_id: str = data[0]
            message_data = {
                FIELD_COMMAND: LpbcastCommand.RETRIEVE_EVENT,
                FIELD_RETRIEVE_EVENT: event_id,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == LpbcastCommand.PUSH_EVENT:
            event = Event(
                id=message.data[FIELD_EVENT_ID],
                data=message.data[FIELD_EVENT_DATA],
                origin=parse_node(message.data[FIELD_EVENT_ORIGIN]),
                digest=parse_digest(message.data[FIELD_DIGEST]),
            )
            return LpbcastCommand.PUSH_EVENT, [event]
        elif command == LpbcastCommand.RETRIEVE_EVENT:
            event_id = message.data[FIELD_RETRIEVE_EVENT]
            return LpbcastCommand.RETRIEVE_EVENT, [event_id]
        raise ValueError("Invalid Command")


class LpbcastProtocol(AProtocol):
    def _get_new_transcoder(self):
        return LpbcastTranscoder(self.my_node, self.service_id)

    async def push_event(self, destination: Node, event: Event):
        message = self._transcoder.encode(LpbcastCommand.PUSH_EVENT, event)
        return await self._rpc.call_send_message(destination, message)

    async def retrieve_event(self, destination: Node, event_id: str):
        message = self._transcoder.encode(LpbcastCommand.RETRIEVE_EVENT, event_id)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_push_event(self, handler: Handler):
        self._handlers[LpbcastCommand.PUSH_EVENT] = handler

    def set_handler_retrieve_event(self, handler: Handler):
        self._handlers[LpbcastCommand.RETRIEVE_EVENT] = handler


def parse_digest(raw_digest: List[Tuple[str, Any]]) -> Digest:
    digest = []
    for message_id, message_origin in raw_digest:
        digest.append((message_id, parse_node(message_origin)))
    return digest



from typing import Any, Callable, List, Optional

from unsserv.common.gossip.gossip import (
    Gossip,
    View,
)
from unsserv.common.services_abc import (
    IClusteringService,
    IMembershipService,
)
from unsserv.common.structs import Node, Property
from unsserv.common.typing import Handler
from unsserv.common.utils import HandlersManager

RankingFunction = Callable[[Node], Any]


class TMan(IClusteringService):
    properties = {Property.EXTREME, Property.HAS_GOSSIP, Property.NON_SYMMETRIC}

    gossip: Optional[Gossip]
    _handler_manager: HandlersManager
    _ranking_function: RankingFunction

    def __init__(self, membership: IMembershipService):
        self.my_node = membership.my_node
        self.membership = membership
        self.gossip = None
        self._handler_manager = HandlersManager()

    async def join(self, service_id: Any, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Clustering")
        self.service_id = service_id
        self._ranking_function = configuration["ranking_function"]  # type: ignore
        local_view_nodes = self.membership.get_neighbours()
        assert isinstance(local_view_nodes, list)  # for mypy validation
        self.gossip = Gossip(
            self.membership.my_node,
            service_id=service_id,
            local_view_nodes=local_view_nodes,
            local_view_handler=self._gossip_local_view_handler,
            custom_selection_ranking=self._selection_ranking,
            external_nodes_source=self.membership.get_neighbours,
        )
        await self.gossip.start()
        self.running = True

    async def leave(self):
        if not self.running:
            return
        await self.gossip.stop()
        self._handler_manager.remove_all_handlers()
        self.gossip = None
        self.running = False

    def get_neighbours(self) -> List[Node]:
        if not self.running:
            raise RuntimeError("Clustering service not running")
        return list(self.gossip.local_view.keys())

    def add_neighbours_handler(self, handler: Handler):
        if not self.running:
            raise RuntimeError("Service not running")
        self._handler_manager.add_handler(handler)

    def remove_neighbours_handler(self, handler: Handler):
        self._handler_manager.remove_handler(handler)

    async def _gossip_local_view_handler(self, local_view: View):
        self._handler_manager.call_handlers(list(local_view.keys()))

    def _selection_ranking(self, view: View) -> List[Node]:
        return sorted(view.keys(), key=self._ranking_function)  # type: ignore



import asyncio
import math
import random
from typing import Dict, List, Any

from unsserv.common.errors import ServiceError
from unsserv.common.services_abc import IMembershipService, ISamplingService
from unsserv.common.structs import Node, Property
from unsserv.common.utils import get_random_id
from unsserv.common.utils import stop_task
from unsserv.extreme.sampling.config import MRWBConfig
from unsserv.extreme.sampling.protocol import MRWBProtocol
from unsserv.extreme.sampling.structs import Sample, SampleResult


class MRWB(ISamplingService):
    properties = {Property.EXTREME}
    _protocol: MRWBProtocol
    _config: MRWBConfig

    _neighbours: List[Node]
    _neighbour_degrees: Dict[Node, int]
    _sampling_queue: Dict[str, Node]
    _sampling_events: Dict[str, asyncio.Event]

    def __init__(self, membership: IMembershipService):
        self.my_node = membership.my_node
        self.membership = membership
        self._protocol = MRWBProtocol(self.my_node)
        self._config = MRWBConfig()

        self._neighbours = []
        self._neighbour_degrees = {}
        self._sampling_queue = {}
        self._sampling_events = {}

    async def join(self, service_id: str, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Sampling")
        self.service_id = service_id
        self._config.load_from_dict(configuration)
        # initialize neighbours
        neighbours = self.membership.get_neighbours()
        assert isinstance(neighbours, list)
        self._neighbours = neighbours
        self._degrees_update_task = asyncio.create_task(
            self._neighbours_degrees_maintenance()
        )  # stop degrees updater task
        # initialize RPC
        await self._initialize_protocol()
        self.membership.add_neighbours_handler(
            self._membership_neighbours_handler  # type: ignore
        )
        self.running = True

    async def leave(self):
        if not self.running:
            return
        self.membership.remove_neighbours_handler(self._membership_neighbours_handler)
        self._neighbours = []
        await self._protocol.stop()
        if self._degrees_update_task:  # stop degrees updater task
            await stop_task(self._degrees_update_task)
        self.running = False

    async def get_sample(self) -> Node:
        if not self.running:
            raise RuntimeError("Sampling service not running")
        sample_id = get_random_id()
        if not self._neighbours:
            raise ServiceError("Unable to peer with neighbours for sampling.")
        event = asyncio.Event()
        self._sampling_events[sample_id] = event
        random_node = self._choose_next_hop()
        sample = Sample(id=sample_id, origin_node=self.my_node, ttl=self._config.TTL)
        await self._protocol.sample(random_node, sample)
        try:
            await asyncio.wait_for(event.wait(), timeout=self._config.TIMEOUT)
        except asyncio.TimeoutError:
            del self._sampling_events[sample_id]
            raise ServiceError("Sampling service timeouted")
        sample_result = self._sampling_queue[sample_id]
        del self._sampling_queue[sample_id]
        return sample_result

    async def _neighbours_degrees_maintenance(self):
        # maybe is not needed if degrees are updated whenever
        # membership changes neighbours?
        while True:
            for neighbour in self._neighbours:
                await self._update_degree(neighbour)
            await asyncio.sleep(self._config.MAINTENANCE_SLEEP)

    def _choose_next_hop(self) -> Node:
        random_neighbour = random.choice(self._neighbours)
        neighbour_degree = self._neighbour_degrees.get(random_neighbour, math.inf)
        my_degree = len(self._neighbours)
        if random.uniform(0, 1) < my_degree / neighbour_degree:
            return random_neighbour
        return self.my_node

    async def _update_degree(self, node: Node):
        try:
            degree = await self._protocol.get_degree(node)
            self._neighbour_degrees[node] = degree
        except ConnectionError:
            pass  # let membership to decide whether to remove the node or not

    async def _membership_neighbours_handler(self, new_neighbours: List[Node]):
        old_neighbours = set(self._neighbours)
        new_neighbours_set = set(new_neighbours)
        for neighbour in old_neighbours - new_neighbours_set:
            if (
                neighbour in self._neighbour_degrees
            ):  # due to concurrency there may be errors
                del self._neighbour_degrees[neighbour]
        for neighbour in new_neighbours_set - old_neighbours:
            await self._update_degree(neighbour)
        self._neighbours = new_neighbours

    async def _handler_sample(self, sender: Node, sample: Sample):
        ttl = sample.ttl
        while ttl > 0:
            next_hop = self._choose_next_hop()
            if next_hop != self.my_node:
                next_sample = Sample(
                    id=sample.id, origin_node=sample.origin_node, ttl=ttl - 1
                )
                asyncio.create_task(self._protocol.sample(next_hop, next_sample))
                return None
            ttl -= 1
        sample_result = SampleResult(sample_id=sample.id, result=self.my_node)
        asyncio.create_task(
            self._protocol.sample_result(sample.origin_node, sample_result)
        )

    async def _handler_sample_result(self, sender: Node, sample_result: SampleResult):
        self._sampling_queue[sample_result.sample_id] = sample_result.result
        self._sampling_events[sample_result.sample_id].set()
        del self._sampling_events[sample_result.sample_id]

    def _handler_get_degree(self, sender: Node):
        return len(self._neighbours)

    async def _initialize_protocol(self):
        self._protocol.set_handler_sample(self._handler_sample)
        self._protocol.set_handler_sample_result(self._handler_sample_result)
        self._protocol.set_handler_get_degree(self._handler_get_degree)
        await self._protocol.start(self.service_id)



import asyncio

import pytest

from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.extreme.membership import newscast
from unsserv.stable.membership import hyparview

first_port = 7771
node = Node(("127.0.0.1", first_port))

NEWSCAST_SERVICE_ID = "newscast"
HYPARVIEW_SERVICE_ID = "hyparview"


def get_random_nodes(amount, first_port=7772, host="127.0.0.1"):
    return list(map(lambda p: Node((host, p)), range(first_port, first_port + amount)))


@pytest.mark.asyncio
@pytest.fixture
async def init_extreme_membership():
    newc = None
    r_newcs = []

    async def get_memberships(amount):
        nonlocal newc, r_newcs
        newc = newscast.Newscast(node)
        await newc.join(NEWSCAST_SERVICE_ID)
        r_nodes = get_random_nodes(amount, first_port=first_port + 1)
        for i, r_node in enumerate(r_nodes):
            r_newc = newscast.Newscast(r_node)
            await r_newc.join(NEWSCAST_SERVICE_ID, bootstrap_nodes=[node] + r_nodes[:i])
            r_newcs.append(r_newc)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return newc, r_newcs

    try:
        yield get_memberships
    finally:
        await newc.leave()
        for r_newc in r_newcs:
            await r_newc.leave()


@pytest.mark.asyncio
@pytest.fixture
async def init_stable_membership():
    hypa = None
    r_hypas = []

    async def get_memberships(amount):
        nonlocal hypa, r_hypas
        hypa = hyparview.HyParView(node)
        await hypa.join(HYPARVIEW_SERVICE_ID)
        r_nodes = get_random_nodes(amount, first_port=first_port + 1)
        for i, r_node in enumerate(r_nodes):
            r_hypa = hyparview.HyParView(r_node)
            await r_hypa.join(
                HYPARVIEW_SERVICE_ID, bootstrap_nodes=[node] + r_nodes[:i]
            )
            r_hypas.append(r_hypa)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return hypa, r_hypas

    try:
        yield get_memberships
    finally:
        await hypa.leave()
        for r_hypa in r_hypas:
            await r_hypa.leave()



from enum import Enum, auto
from typing import Dict, Any

from unsserv.common.utils import IConfig


class ViewSelectionPolicy(Enum):
    RAND = auto()
    HEAD = auto()
    TAIL = auto()


class PeerSelectionPolicy(Enum):
    RAND = auto()
    HEAD = auto()
    TAIL = auto()


class ViewPropagationPolicy(Enum):
    PUSH = auto()
    PULL = auto()
    PUSHPULL = auto()


class GossipConfig(IConfig):
    LOCAL_VIEW_SIZE = 10  # todo: select a proper size
    GOSSIPING_FREQUENCY = 0.2  # todo: select a proper size
    VIEW_SELECTION = ViewSelectionPolicy.HEAD
    VIEW_PROPAGATION = ViewPropagationPolicy.PUSHPULL
    PEER_SELECTION = PeerSelectionPolicy.RAND
    RPC_TIMEOUT = 1

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.LOCAL_VIEW_SIZE = config_dict.get(
            "local_view_size", GossipConfig.LOCAL_VIEW_SIZE
        )
        self.GOSSIPING_FREQUENCY = config_dict.get(
            "gossiping_frequency", GossipConfig.GOSSIPING_FREQUENCY
        )
        self.VIEW_SELECTION = config_dict.get(
            "view_selection", GossipConfig.VIEW_SELECTION
        )
        self.VIEW_PROPAGATION = config_dict.get(
            "view_propagation", GossipConfig.VIEW_PROPAGATION
        )
        self.PEER_SELECTION = config_dict.get(
            "peer_selection", GossipConfig.PEER_SELECTION
        )
        self.RPC_TIMEOUT = config_dict.get("rpc_timeout", GossipConfig.RPC_TIMEOUT)



from typing import Dict, Any

from unsserv.common.utils import IConfig


class KWalkerConfig(IConfig):
    TTL = 4
    FANOUT = 4
    TIMEOUT = 4

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.TTL = config_dict.get("ttl", KWalkerConfig.TTL)
        self.FANOUT = config_dict.get("fanout", KWalkerConfig.FANOUT)
        self.TIMEOUT = config_dict.get("timeout", KWalkerConfig.TIMEOUT)



from dataclasses import dataclass
from typing import Dict, Any

from unsserv.common.gossip.typing import Payload, View


@dataclass
class PushData:
    view: View
    payload: Payload

    def encode(self) -> Dict[str, Any]:
        return {"view": self.view, "payload": self.payload}



from dataclasses import dataclass

from unsserv.common.structs import Node


@dataclass
class Search:
    id: str
    origin_node: Node
    ttl: int
    data_id: str


@dataclass
class SearchResult:
    search_id: str
    result: bytes


@dataclass
class DataChange:
    data_id: str
    ttl: int



import asyncio
from collections import Counter
from functools import partial
from math import ceil

import pytest

from tests.utils import init_extreme_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.extreme.clustering.t_man import TMan

init_extreme_membership = init_extreme_membership  # for flake8 compliance

CLUSTERING_SERVICE_ID = "tman"


@pytest.mark.asyncio
@pytest.fixture
async def init_tman():
    tman = None
    r_tmans = []

    async def _init_tman(newc, r_newcs):
        nonlocal tman, r_tmans
        tman = TMan(newc)
        await tman.join(
            CLUSTERING_SERVICE_ID, ranking_function=partial(port_distance, tman.my_node)
        )
        for r_newc in r_newcs:
            r_tman = TMan(r_newc)
            await r_tman.join(
                CLUSTERING_SERVICE_ID,
                ranking_function=partial(port_distance, r_tman.my_node),
            )
            r_tmans.append(r_tman)
        return tman, r_tmans

    try:
        yield _init_tman
    finally:
        await tman.leave()
        for r_tman in r_tmans:
            await r_tman.leave()


def port_distance(my_node: Node, ranked_node: Node):
    return abs(my_node.address_info[1] - ranked_node.address_info[1])


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [GossipConfig.LOCAL_VIEW_SIZE * 2 + 1, GossipConfig.LOCAL_VIEW_SIZE * 2 + 5, 100],
)
async def test_join_tman(init_extreme_membership, init_tman, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    tman, r_tmans = await init_tman(newc, r_newcs)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 40)

    cluster_nodes = [tman] + r_tmans
    satisfy_ideal_neighbours = []
    for cluster in cluster_nodes:
        neighbours = set(cluster.get_neighbours())
        key_function = partial(port_distance, cluster.my_node)
        ideal_neighbours = set(
            sorted(map(lambda c_n: c_n.my_node, cluster_nodes), key=key_function)[
                1 : GossipConfig.LOCAL_VIEW_SIZE + 1
            ]
        )
        satisfies_half_ideal_neighbours = min(
            amount, GossipConfig.LOCAL_VIEW_SIZE
        ) * 0.5 <= len(ideal_neighbours.intersection(neighbours))
        satisfy_ideal_neighbours.append(satisfies_half_ideal_neighbours)
    assert sum(satisfy_ideal_neighbours) / (amount + 1) >= 0.5


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        GossipConfig.LOCAL_VIEW_SIZE + 1,
        GossipConfig.LOCAL_VIEW_SIZE + 5,
        GossipConfig.LOCAL_VIEW_SIZE + 100,
    ],
)
async def test_leave_tman(init_extreme_membership, init_tman, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    tman, r_tmans = await init_tman(newc, r_newcs)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)

    await tman.leave()
    await newc.leave()
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 40)

    all_nodes = Counter(
        [
            item
            for sublist in map(lambda n: n.get_neighbours(), r_tmans)
            for item in sublist
        ]
    )
    nodes_ten_percent = ceil(amount * 0.2)
    assert newc.my_node not in all_nodes.keys() or newc.my_node in set(
        map(lambda p: p[0], all_nodes.most_common()[-nodes_ten_percent:])
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 1,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 5,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 100,
    ],
)  # very high neighbours amount,
# to assure neighbours will change, because it is initailzied by Newscast
async def test_tman_handler(init_extreme_membership, init_tman, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    tman, r_tmans = await init_tman(newc, r_newcs)

    handler_event = asyncio.Event()

    async def handler(local_view):
        assert isinstance(local_view, list)
        nonlocal handler_event
        handler_event.set()

    tman.add_neighbours_handler(handler)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)
    assert handler_event.is_set()



import asyncio
from functools import partial
from typing import Any, Dict

import pytest

from tests.utils import init_extreme_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.extreme.dissemination.mon.mon import Mon

init_extreme_membership = init_extreme_membership  # for flake8 compliance

DISSEMINATION_SERVICE_ID = "mon"

mon_events: Dict[Node, asyncio.Event] = {}


async def dissemination_handler(node: Node, data: Any):
    mon_events[node].set()


@pytest.mark.asyncio
@pytest.fixture
async def init_mon():
    mon = None
    r_mons = []

    async def _init_mon(newc, r_newcs):
        nonlocal mon, r_mons
        mon = Mon(newc)
        mon_events[newc.my_node] = asyncio.Event()
        await mon.join(
            DISSEMINATION_SERVICE_ID,
            broadcast_handler=partial(dissemination_handler, mon.my_node),
        )
        for r_newc in r_newcs:
            r_mon = Mon(r_newc)
            mon_events[r_newc.my_node] = asyncio.Event()
            await r_mon.join(
                DISSEMINATION_SERVICE_ID,
                broadcast_handler=partial(dissemination_handler, r_mon.my_node),
            )
            r_mons.append(r_mon)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return mon, r_mons

    try:
        yield _init_mon
    finally:
        await mon.leave()
        for r_mon in r_mons:
            await r_mon.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_extreme_membership, init_mon, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    mon, r_mons = await init_mon(newc, r_newcs)


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_broadcast(init_extreme_membership, init_mon, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    mon, r_mons = await init_mon(newc, r_newcs)

    data = b"data"
    await mon.broadcast(data)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)

    mon_events_received = [mon_events[r_mon.my_node].is_set() for r_mon in r_mons]
    assert int(amount * 0.75) <= sum(mon_events_received)

    await mon.leave()
    for r_mon in r_mons:
        await r_mon.leave()



import asyncio

import pytest

from tests.utils import init_extreme_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.extreme.sampling.mrwb import MRWB

init_extreme_membership = init_extreme_membership  # for flake8 compliance

AGGREGATION_SERVICE_ID = "mrwb"


@pytest.mark.asyncio
@pytest.fixture
async def init_mrwb():
    sampl = None
    r_sampls = []

    async def _init_mrwb(newc, r_newcs):
        nonlocal sampl, r_sampls
        sampl = MRWB(newc)
        await sampl.join(AGGREGATION_SERVICE_ID)
        for r_newc in r_newcs:
            r_sampl = MRWB(r_newc)
            await r_sampl.join(AGGREGATION_SERVICE_ID)
            r_sampls.append(r_sampl)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return r_sampls, sampl

    try:
        yield _init_mrwb
    finally:
        await sampl.leave()
        for r_sampl in r_sampls:
            await r_sampl.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_extreme_membership, init_mrwb, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    r_sampls, sampl = await init_mrwb(newc, r_newcs)


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_sampling(init_extreme_membership, init_mrwb, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    r_sampls, sampl = await init_mrwb(newc, r_newcs)

    samples = {await sampl.get_sample() for _ in range(amount * 2)}
    assert len({r_newc.my_node for r_newc in r_newcs} - samples) / len(r_newcs) <= 0.45



import asyncio
import random
from abc import ABC, abstractmethod
from collections import Counter
from contextlib import contextmanager
from typing import List, Set, Counter as CounterType

from unsserv.common.structs import Node
from unsserv.common.utils import stop_task, HandlersManager
from unsserv.stable.membership.double_layered.config import DoubleLayeredConfig
from unsserv.stable.membership.double_layered.protocol import DoubleLayeredProtocol
from unsserv.stable.membership.double_layered.structs import ForwardJoin


class IDoubleLayered(ABC):
    _handler_manager: HandlersManager
    _doble_layered_protocol: DoubleLayeredProtocol
    _config: DoubleLayeredConfig

    _active_view: Set[Node]
    _candidate_neighbours: CounterType[Node]
    _local_view_maintenance_task: asyncio.Task

    def __init__(self, my_node: Node):
        self.my_node = my_node
        self._handler_manager = HandlersManager()
        self._doble_layered_protocol = DoubleLayeredProtocol(my_node)
        self._active_view = set()
        self._candidate_neighbours = Counter()

    async def _start_two_layered(self, service_id: str):
        await self._initialize_double_layered_protocol(service_id)
        self._local_view_maintenance_task = asyncio.create_task(
            self._maintain_active_view_loop()
        )

    async def _stop_two_layered(self):
        self._active_view = set()
        if self._local_view_maintenance_task:
            await stop_task(self._local_view_maintenance_task)
        await self._doble_layered_protocol.stop()

    async def _join_first_time(self):
        bootstrap_nodes = self._get_passive_view_nodes()
        while (
            bootstrap_nodes
            and len(self._active_view) + len(self._candidate_neighbours)
            < self._config.ACTIVE_VIEW_SIZE
        ):
            candidate_node = bootstrap_nodes.pop(random.randrange(len(bootstrap_nodes)))
            with self._create_candidate_neighbour(candidate_node):
                try:
                    await self._doble_layered_protocol.join(candidate_node)
                    self._active_view.add(candidate_node)
                    break
                except ConnectionError:
                    pass

    async def _maintain_active_view_loop(self):
        await self._join_first_time()
        while True:
            old_active_view = self._active_view.copy()
            await asyncio.sleep(self._config.MAINTENANCE_SLEEP)
            await self._update_active_view()
            self._call_handler_if_view_changed(old_active_view)

    async def _update_active_view(self):
        inactive_nodes = set()
        for node in self._active_view.copy():
            try:
                is_still_connected = await self._doble_layered_protocol.stay_connected(
                    node
                )
                if not is_still_connected:
                    inactive_nodes.add(node)
            except ConnectionError:
                inactive_nodes.add(node)
        self._active_view = self._active_view - inactive_nodes
        if len(self._active_view) >= self._config.ACTIVE_VIEW_SIZE:
            return
        candidate_neighbours: List[Node] = self._get_passive_view_nodes()
        while (
            candidate_neighbours
            and len(self._active_view) + len(self._candidate_neighbours)
            < self._config.ACTIVE_VIEW_SIZE
        ):
            candidate_neighbour = candidate_neighbours.pop(
                random.randrange(len(candidate_neighbours))
            )
            await self._connect_to_node(candidate_neighbour)

    async def _connect_to_node(self, node: Node):
        with self._create_candidate_neighbour(node):
            is_a_priority = len(self._active_view) == 0
            try:
                is_connected = await self._doble_layered_protocol.connect(
                    node, is_a_priority
                )
                if is_connected:
                    self._force_add_neighbour(node)
            except ConnectionError:
                pass

    async def _try_disconnect(self, node: Node):
        try:
            await self._doble_layered_protocol.disconnect(node)
        except ConnectionError:
            pass

    def _force_add_neighbour(self, node: Node):
        while len(self._active_view) >= self._config.ACTIVE_VIEW_SIZE:
            random_neighbour = random.choice(list(self._active_view))
            self._active_view.remove(random_neighbour)  # randomly remove
            asyncio.create_task(self._try_disconnect(random_neighbour))
        self._active_view.add(node)

    def _call_handler_if_view_changed(self, old_local_view: Set):
        if old_local_view == self._active_view:
            return
        self._handler_manager.call_handlers(list(self._active_view))

    @contextmanager
    def _create_candidate_neighbour(self, node: Node):
        self._candidate_neighbours.update([node])
        try:
            yield
        finally:
            self._candidate_neighbours.subtract([node])
            self._candidate_neighbours = (
                +self._candidate_neighbours
            )  # remove zero and negative counts

    async def _handler_join(self, sender: Node):
        self._force_add_neighbour(sender)
        forward_join = ForwardJoin(origin_node=sender, ttl=self._config.TTL)
        for neighbour in list(filter(lambda n: n != sender, self._active_view)):
            asyncio.create_task(
                self._doble_layered_protocol.forward_join(neighbour, forward_join)
            )

    async def _handler_forward_join(self, sender: Node, forward_join: ForwardJoin):
        if forward_join.ttl == 0:
            asyncio.create_task(self._connect_to_node(forward_join.origin_node))
        else:
            candidate_neighbours = list(
                filter(lambda n: n != forward_join.origin_node, self._active_view)
            ) or [self.my_node]
            neighbour = random.choice(candidate_neighbours)
            next_forward_join = ForwardJoin(
                origin_node=forward_join.origin_node, ttl=forward_join.ttl - 1
            )
            asyncio.create_task(
                self._doble_layered_protocol.forward_join(neighbour, next_forward_join)
            )

    async def _handler_connect(self, sender: Node, is_a_priority: bool):
        if (
            not is_a_priority
            and len(self._active_view) >= self._config.ACTIVE_VIEW_SIZE
        ):
            return False
        self._force_add_neighbour(sender)
        return True

    async def _handler_disconnect(self, sender: Node):
        if sender in self._active_view:
            self._active_view.remove(sender)

    async def _handler_stay_connected(self, sender: Node):
        return sender in self._active_view or sender in self._candidate_neighbours

    async def _initialize_double_layered_protocol(self, service_id: str):
        self._doble_layered_protocol.set_handler_join(self._handler_join)
        self._doble_layered_protocol.set_handler_forward_join(
            self._handler_forward_join
        )
        self._doble_layered_protocol.set_handler_connect(self._handler_connect)
        self._doble_layered_protocol.set_handler_disconnect(self._handler_disconnect)
        self._doble_layered_protocol.set_handler_stay_connected(
            self._handler_stay_connected
        )
        await self._doble_layered_protocol.start(service_id)

    @abstractmethod
    def _get_passive_view_nodes(self):
        pass



from enum import IntEnum, auto
from typing import Tuple, Sequence

from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.extreme.dissemination.mon.structs import Session, Broadcast

FIELD_COMMAND = "mon-command"
FIELD_BROADCAST_ID = "mon-broadcast-id"
FIELD_LEVEL = "mon-level"
FIELD_BROADCAST_DATA = "mon-broadcast-data"


class MonCommand(IntEnum):
    SESSION = auto()
    PUSH = auto()


class MonTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == MonCommand.SESSION:
            session: Session = data[0]
            message_data = {
                FIELD_COMMAND: MonCommand.SESSION,
                FIELD_BROADCAST_ID: session.broadcast_id,
                FIELD_LEVEL: session.level,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == MonCommand.PUSH:
            broadcast: Broadcast = data[0]
            message_data = {
                FIELD_COMMAND: MonCommand.PUSH,
                FIELD_BROADCAST_ID: broadcast.id,
                FIELD_BROADCAST_DATA: broadcast.data,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == MonCommand.SESSION:
            session = Session(
                broadcast_id=message.data[FIELD_BROADCAST_ID],
                level=message.data[FIELD_LEVEL],
            )
            return MonCommand.SESSION, [session]
        elif command == MonCommand.PUSH:
            broadcast = Broadcast(
                id=message.data[FIELD_BROADCAST_ID],
                data=message.data[FIELD_BROADCAST_DATA],
            )
            return MonCommand.PUSH, [broadcast]
        raise ValueError("Invalid Command")


class MonProtocol(AProtocol):
    def _get_new_transcoder(self):
        return MonTranscoder(self.my_node, self.service_id)

    async def session(self, destination: Node, session: Session) -> bool:
        message = self._transcoder.encode(MonCommand.SESSION, session)
        return await self._rpc.call_send_message(destination, message)

    async def push(self, destination: Node, broadcast: Broadcast):
        message = self._transcoder.encode(MonCommand.PUSH, broadcast)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_session(self, handler: Handler):
        self._handlers[MonCommand.SESSION] = handler

    def set_handler_push(self, handler: Handler):
        self._handlers[MonCommand.PUSH] = handler



import asyncio
from contextlib import asynccontextmanager
from functools import reduce
from typing import Optional, Any, List, Dict, Set

from unsserv.common.services_abc import ISearchingService, IMembershipService
from unsserv.common.structs import Node, Property
from unsserv.common.utils import get_random_id, stop_task
from unsserv.stable.searching.config import ABloomConfig
from unsserv.stable.searching.protocol import ABloomProtocol
from unsserv.stable.searching.structs import Search, SearchResult, DataChange
from unsserv.stable.searching.typing import DataID, SearchID


class ABloom(ISearchingService):
    properties = {Property.STABLE}
    _protocol: ABloomProtocol
    _config: ABloomConfig

    _neighbours: List[Node]
    _local_data: Dict[DataID, bytes]
    _abloom_filters: Dict[Node, List[Set[DataID]]]
    _filters_maintenance_task: asyncio.Task
    _search_events: Dict[SearchID, asyncio.Event]
    _search_results: Dict[SearchID, bytes]

    def __init__(self, membership: IMembershipService):
        self.my_node = membership.my_node
        self.membership = membership
        self._protocol = ABloomProtocol(self.my_node)
        self._config = ABloomConfig()

        self._init_structs()

    async def join(self, service_id: str, **configuration: Any):
        if self.running:
            raise RuntimeError("Service already running")
        self.service_id = service_id
        await self._initialize_protocol()
        self._config.load_from_dict(configuration)
        self._init_structs()
        self._filters_maintenance_task = asyncio.create_task(
            self._filter_maintenance_loop()
        )
        self.running = True

    async def leave(self):
        if not self.running:
            return
        await stop_task(self._filters_maintenance_task)
        await self._protocol.stop()
        self.running = False

    async def publish(self, data_id: str, data: bytes):
        if not self.running:
            raise RuntimeError("Service must be running for pubslishing data")
        assert isinstance(data, bytes)
        if data_id in self._local_data:
            raise KeyError("Duplicated Key")
        self._local_data[data_id] = data
        asyncio.create_task(self._publish_to_neighbours(data_id))

    async def unpublish(self, data_id: str):
        if not self.running:
            raise RuntimeError("Service must be running for unpubslishing data")
        del self._local_data[data_id]
        asyncio.create_task(self._unpublish_to_neighbours(data_id))

    async def search(self, data_id: str) -> Optional[bytes]:
        if data_id in self._local_data:
            return self._local_data[data_id]
        next_hop = self._search_data_in_filters(data_id, self._config.DEPTH)
        if not next_hop:
            return None
        async with self._start_search(data_id, next_hop) as search:
            await asyncio.wait_for(
                self._search_events[search.id].wait(), timeout=self._config.TIMEOUT
            )
            return self._search_results[search.id]

    async def _publish_to_neighbours(self, data_id: DataID):
        data_change = DataChange(data_id=data_id, ttl=self._config.DEPTH)
        for neighbour in self._neighbours:
            asyncio.create_task(self._protocol.publish(neighbour, data_change))

    async def _unpublish_to_neighbours(self, data_id: DataID):
        data_change = DataChange(data_id=data_id, ttl=self._config.DEPTH)
        for neighbour in self._neighbours:
            asyncio.create_task(self._protocol.unpublish(neighbour, data_change))

    @asynccontextmanager
    async def _start_search(self, data_id: DataID, next_hop: Node):
        search = Search(
            id=get_random_id(),
            origin_node=self.my_node,
            ttl=self._config.DEPTH,
            data_id=data_id,
        )
        self._search_events[search.id] = asyncio.Event()
        try:
            await self._protocol.search(next_hop, search)
            yield search
        finally:
            del self._search_events[search.id]
            self._remove_id_from_filter(data_id, next_hop)
            if data_id in self._search_results:
                del self._search_results[data_id]

    def _init_structs(self):
        self._neighbours = []
        self._abloom_filters = {}
        self._abloom_filters = {}
        self._local_data = {}

        self._search_events = {}
        self._search_results = {}

    async def _initialize_new_neighbours(self, new_neighbours: List[Node]):
        for neighbour in new_neighbours:
            try:
                self._abloom_filters[neighbour] = await self._protocol.get_filter(
                    neighbour
                )
                self._neighbours.append(neighbour)
            except ConnectionError:
                pass

    async def _filter_maintenance_loop(self):
        while True:
            await asyncio.sleep(self._config.MAINTENANCE_SLEEP)
            new_neighbours = set(self.membership.get_neighbours())
            old_neighbours = set(self._neighbours)
            for neighbour in old_neighbours - new_neighbours:
                del self._abloom_filters[neighbour]
                self._neighbours.remove(neighbour)
            await self._initialize_new_neighbours(list(new_neighbours - old_neighbours))

    def _search_data_in_filters(self, data_id: str, max_depth: int) -> Optional[Node]:
        for neighbour in self._neighbours:
            for depth in range(max_depth):
                if data_id in self._abloom_filters[neighbour][depth]:
                    return neighbour
        return None

    def _remove_id_from_filter(self, data_id: DataID, neighbour: Node):
        for depth in range(self._config.DEPTH):
            if data_id in self._abloom_filters[neighbour][depth]:
                self._abloom_filters[neighbour][depth].remove(data_id)
                return

    async def _handler_publish(self, sender: Node, data_change: DataChange):
        if sender not in self._neighbours:
            return
        depth = self._config.DEPTH - data_change.ttl
        self._abloom_filters[sender][depth].add(data_change.data_id)
        if data_change.ttl < 2:
            return
        next_data_change = DataChange(
            data_id=data_change.data_id, ttl=data_change.ttl - 1
        )
        for neighbour in self._neighbours:
            asyncio.create_task(self._protocol.publish(neighbour, next_data_change))

    async def _handler_unpublish(self, sender: Node, data_change: DataChange):
        if sender not in self._neighbours:
            return
        depth = self._config.DEPTH - data_change.ttl
        self._abloom_filters[sender][depth].remove(data_change.data_id)
        if data_change.ttl < 2:
            return
        next_data_change = DataChange(
            data_id=data_change.data_id, ttl=data_change.ttl - 1
        )
        for neighbour in self._neighbours:
            asyncio.create_task(self._protocol.unpublish(neighbour, next_data_change))

    async def _handler_get_filter(self, sender: Node):
        filter = [set(self._local_data.keys())]
        for depth in range(1, self._config.DEPTH):
            filters = list(
                map(lambda n: self._abloom_filters[n][depth - 1], self._neighbours)
            )
            if filters:
                filter.append(reduce(lambda f1, f2: f1.union(f2), filters))
            else:
                filter.append(set())
        return filter

    async def _handler_search(self, sender: Node, search: Search):
        if search.data_id in self._local_data:
            search_result = SearchResult(
                search_id=search.id, result=self._local_data[search.data_id]
            )
            asyncio.create_task(
                self._protocol.search_result(search.origin_node, search_result)
            )
            return
        next_hop = self._search_data_in_filters(search.data_id, search.ttl - 1)
        if next_hop and search.ttl > 1:
            next_search = Search(
                id=search.id,
                origin_node=search.origin_node,
                ttl=search.ttl - 1,
                data_id=search.data_id,
            )
            asyncio.create_task(self._protocol.search(next_hop, next_search))
        else:
            raise ValueError("Data was not found")

    async def _handler_search_result(self, sender: Node, search_result: SearchResult):
        self._search_results[search_result.search_id] = search_result.result
        self._search_events[search_result.search_id].set()

    async def _initialize_protocol(self):
        self._protocol.set_handler_publish(self._handler_publish)
        self._protocol.set_handler_unpublish(self._handler_unpublish)
        self._protocol.set_handler_get_filter(self._handler_get_filter)
        self._protocol.set_handler_search(self._handler_search)
        self._protocol.set_handler_search_result(self._handler_search_result)
        await self._protocol.start(self.service_id)



import asyncio
from collections import Counter
from functools import partial
from math import ceil

import pytest

from tests.utils import init_extreme_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.stable.clustering.config import XBotConfig
from unsserv.stable.clustering.x_bot import XBot

init_extreme_membership = init_extreme_membership  # for flake8 compliance

CLUSTERING_SERVICE_ID = "xbot"


@pytest.mark.asyncio
@pytest.fixture
async def init_xbot():
    xbot = None
    r_xbots = []

    async def _init_xbot(newc, r_newcs):
        nonlocal xbot, r_xbots
        xbot = XBot(newc)
        await xbot.join(
            CLUSTERING_SERVICE_ID, ranking_function=partial(port_distance, xbot.my_node)
        )
        for r_newc in r_newcs:
            r_xbot = XBot(r_newc)
            await r_xbot.join(
                CLUSTERING_SERVICE_ID,
                ranking_function=partial(port_distance, r_xbot.my_node),
            )
            r_xbots.append(r_xbot)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return xbot, r_xbots

    try:
        yield _init_xbot
    finally:
        await xbot.leave()
        for r_xbot in r_xbots:
            await r_xbot.leave()


def port_distance(my_node: Node, ranked_node: Node):
    return abs(my_node.address_info[1] - ranked_node.address_info[1])


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [GossipConfig.LOCAL_VIEW_SIZE * 2 + 1, GossipConfig.LOCAL_VIEW_SIZE * 2 + 5, 100],
)
async def test_join_xbot(init_extreme_membership, init_xbot, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    xbot, r_xbots = await init_xbot(newc, r_newcs)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 40)

    cluster_nodes = [xbot] + r_xbots
    satisfy_ideal_neighbours = []
    for cluster in cluster_nodes:
        neighbours = set(cluster.get_neighbours()[XBotConfig.UNBIASED_NODES :])
        key_function = partial(port_distance, cluster.my_node)
        ideal_neighbours = set(
            sorted(map(lambda c_n: c_n.my_node, cluster_nodes), key=key_function)[
                1 : (XBotConfig.ACTIVE_VIEW_SIZE - XBotConfig.UNBIASED_NODES) + 1
            ]
        )
        satisfies_half_ideal_neighbours = min(
            amount, (XBotConfig.ACTIVE_VIEW_SIZE - XBotConfig.UNBIASED_NODES)
        ) * 0.15 <= len(ideal_neighbours.intersection(neighbours))
        satisfy_ideal_neighbours.append(satisfies_half_ideal_neighbours)
    assert sum(satisfy_ideal_neighbours) / (amount + 1) >= 0.5


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        GossipConfig.LOCAL_VIEW_SIZE + 1,
        GossipConfig.LOCAL_VIEW_SIZE + 5,
        GossipConfig.LOCAL_VIEW_SIZE + 100,
    ],
)
async def test_leave_xbot(init_extreme_membership, init_xbot, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    xbot, r_xbots = await init_xbot(newc, r_newcs)

    await xbot.leave()
    await newc.leave()
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 40)

    all_nodes = Counter(
        [
            item
            for sublist in map(lambda n: n.get_neighbours(), r_xbots)
            for item in sublist
        ]
    )
    nodes_ten_percent = ceil(amount * 0.2)
    assert newc.my_node not in all_nodes.keys() or newc.my_node in set(
        map(lambda p: p[0], all_nodes.most_common()[-nodes_ten_percent:])
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 1,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 5,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 100,
    ],
)  # very high neighbours amount,
# to assure neighbours will change, because it is initailzied by Newscast
async def test_xbot_handler(init_extreme_membership, init_xbot, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    xbot, r_xbots = await init_xbot(newc, r_newcs)

    handler_event = asyncio.Event()

    async def handler(local_view):
        assert isinstance(local_view, list)
        nonlocal handler_event
        handler_event.set()

    xbot.add_neighbours_handler(handler)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)
    assert handler_event.is_set()



import asyncio
import random
import string
from abc import ABC, abstractmethod
from typing import List, Dict, Any

from unsserv.common.rpc.structs import Message
from unsserv.common.structs import Node
from unsserv.common.typing import Handler, SyncHandler


def parse_node(raw_node: List) -> Node:
    address_info = tuple(raw_node[0])
    extra = tuple(raw_node[1])
    return Node(address_info, extra)


def parse_message(raw_message: List) -> Message:
    node = parse_node(raw_message[0])
    return Message(node, raw_message[1], raw_message[2])


def get_random_id(size: int = 10) -> str:
    id_characters = string.ascii_letters + string.digits + string.punctuation
    return "".join(random.choice(id_characters) for _ in range(size))


async def stop_task(task: asyncio.Task):
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


class HandlersManager:
    _handlers: List[Handler]

    def __init__(self):
        self._handlers = []

    def add_handler(self, handler: Handler):
        self._handlers.append(handler)

    def remove_handler(self, handler: Handler):
        self._handlers.remove(handler)

    def remove_all_handlers(self):
        self._handlers.clear()

    def call_handlers(self, *args, **kwargs):
        for handler in self._handlers:
            if asyncio.iscoroutinefunction(handler):
                asyncio.create_task(handler(*args, **kwargs))
            elif hasattr(handler, "func") and asyncio.iscoroutinefunction(
                handler.func
            ):  # in case 'partial' was used
                asyncio.create_task(handler(*args, **kwargs))
            else:
                asyncio.create_task(
                    self._sync_handler_wrapper(handler, *args, **kwargs)
                )

    async def _sync_handler_wrapper(self, sync_handler: SyncHandler, *args, **kwargs):
        sync_handler(*args, **kwargs)


class IConfig(ABC):
    @abstractmethod
    def load_from_dict(self, config_dict: Dict[str, Any]):
        pass



from enum import IntEnum, auto
from typing import Tuple, Sequence

from unsserv.common.utils import parse_node
from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.extreme.sampling.structs import Sample, SampleResult

FIELD_COMMAND = "mrwb-command"
FIELD_TTL = "mrwb-ttl"
FIELD_ORIGIN_NODE = "mrwb-origin-node"
FIELD_SAMPLE_RESULT = "mrwb-sample-result"
FIELD_SAMPLE_ID = "mrwb-sample-id"


class MRWBCommand(IntEnum):
    GET_DEGREE = auto()
    SAMPLE = auto()
    SAMPLE_RESULT = auto()


class MRWBTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == MRWBCommand.SAMPLE:
            sample: Sample = data[0]
            message_data = {
                FIELD_COMMAND: MRWBCommand.SAMPLE,
                FIELD_SAMPLE_ID: sample.id,
                FIELD_ORIGIN_NODE: sample.origin_node,
                FIELD_TTL: sample.ttl,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == MRWBCommand.SAMPLE_RESULT:
            sample_result: SampleResult = data[0]
            message_data = {
                FIELD_COMMAND: MRWBCommand.SAMPLE_RESULT,
                FIELD_SAMPLE_ID: sample_result.sample_id,
                FIELD_SAMPLE_RESULT: sample_result.result,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == MRWBCommand.GET_DEGREE:
            message_data = {FIELD_COMMAND: MRWBCommand.GET_DEGREE}
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == MRWBCommand.SAMPLE:
            sample = Sample(
                id=message.data[FIELD_SAMPLE_ID],
                origin_node=parse_node(message.data[FIELD_ORIGIN_NODE]),
                ttl=message.data[FIELD_TTL],
            )
            return MRWBCommand.SAMPLE, [sample]
        elif command == MRWBCommand.SAMPLE_RESULT:
            sample_result = SampleResult(
                sample_id=message.data[FIELD_SAMPLE_ID],
                result=parse_node(message.data[FIELD_SAMPLE_RESULT]),
            )
            return MRWBCommand.SAMPLE_RESULT, [sample_result]
        elif command == MRWBCommand.GET_DEGREE:
            return MRWBCommand.GET_DEGREE, []
        raise ValueError("Invalid Command")


class MRWBProtocol(AProtocol):
    def _get_new_transcoder(self):
        return MRWBTranscoder(self.my_node, self.service_id)

    async def sample(self, destination: Node, sample: Sample):
        message = self._transcoder.encode(MRWBCommand.SAMPLE, sample)
        return await self._rpc.call_send_message(destination, message)

    async def sample_result(self, destination: Node, sample_result: SampleResult):
        message = self._transcoder.encode(MRWBCommand.SAMPLE_RESULT, sample_result)
        return await self._rpc.call_send_message(destination, message)

    async def get_degree(self, destination: Node) -> int:
        message = self._transcoder.encode(MRWBCommand.GET_DEGREE)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_sample(self, handler: Handler):
        self._handlers[MRWBCommand.SAMPLE] = handler

    def set_handler_sample_result(self, handler: Handler):
        self._handlers[MRWBCommand.SAMPLE_RESULT] = handler

    def set_handler_get_degree(self, handler: Handler):
        self._handlers[MRWBCommand.GET_DEGREE] = handler



from typing import Any, List, Optional

from unsserv.common.gossip.gossip import Gossip, View
from unsserv.common.services_abc import IMembershipService
from unsserv.common.structs import Node, Property
from unsserv.common.typing import Handler
from unsserv.common.utils import HandlersManager


class Newscast(IMembershipService):
    properties = {Property.EXTREME, Property.HAS_GOSSIP, Property.NON_SYMMETRIC}
    gossip: Optional[Gossip]
    _handler_manager: HandlersManager

    def __init__(self, node: Node):
        self.my_node = node
        self.gossip = None
        self._handler_manager = HandlersManager()

    async def join(self, service_id: Any, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Membership")
        self.service_id = service_id
        self.gossip = Gossip(
            my_node=self.my_node,
            service_id=service_id,
            local_view_nodes=configuration.get("bootstrap_nodes", None),
            local_view_handler=self._gossip_local_view_handler,
        )
        await self.gossip.start()
        self.running = True

    async def leave(self):
        if not self.running:
            return
        await self.gossip.stop()
        self.gossip = None
        self.running = False

    def get_neighbours(self) -> List[Node]:
        if not self.running:
            raise RuntimeError("Membership service not running")
        return list(self.gossip.local_view.keys())

    def add_neighbours_handler(self, handler: Handler):
        self._handler_manager.add_handler(handler)

    def remove_neighbours_handler(self, handler: Handler):
        self._handler_manager.remove_handler(handler)

    async def _gossip_local_view_handler(self, local_view: View):
        self._handler_manager.call_handlers(list(local_view.keys()))



from typing import Dict, Any

from unsserv.common.utils import IConfig


class MonConfig(IConfig):
    TIMEOUT = 5  # seconds
    FANOUT = 10
    TREE_LIFE_TIME = 10

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.TIMEOUT = config_dict.get("timeout", MonConfig.TIMEOUT)
        self.FANOUT = config_dict.get("fanout", MonConfig.FANOUT)
        self.TREE_LIFE_TIME = config_dict.get(
            "tree_life_timeout", MonConfig.TREE_LIFE_TIME
        )



from typing import Dict, Any

from unsserv.common.utils import IConfig


class MRWBConfig(IConfig):
    MAINTENANCE_SLEEP = 0.5
    TIMEOUT = 2
    TTL = 10

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.MAINTENANCE_SLEEP = config_dict.get(
            "MAINTENANCE_SLEEP", MRWBConfig.MAINTENANCE_SLEEP
        )
        self.TIMEOUT = config_dict.get("timeout", MRWBConfig.TIMEOUT)
        self.TTL = config_dict.get("ttl", MRWBConfig.TTL)



from dataclasses import dataclass
from unsserv.common.structs import Node


@dataclass
class Walk:
    id: str
    data_id: str
    origin_node: Node
    ttl: int


@dataclass
class WalkResult:
    walk_id: str
    result: bytes



from enum import Enum, auto
from typing import Dict, Any

from unsserv.common.utils import IConfig


class AggregateType(Enum):
    MEAN = auto()
    MAX = auto()
    MIN = auto()


class AntiConfig(IConfig):
    AGGREGATE_TYPE = None

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.AGGREGATE_TYPE = config_dict["aggregate_type"]



from typing import Dict, Any

from unsserv.common.utils import IConfig


class BrisaConfig(IConfig):
    TIMEOUT = 5  # seconds
    FANOUT = 3
    MAINTENANCE_SLEEP = 0.5

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.TIMEOUT = config_dict.get("timeout", BrisaConfig.TIMEOUT)
        self.FANOUT = config_dict.get("fanout", BrisaConfig.FANOUT)
        self.MAINTENANCE_SLEEP = config_dict.get(
            "maintenance_sleep", BrisaConfig.MAINTENANCE_SLEEP
        )



from collections import namedtuple
from enum import Enum, auto

Node = namedtuple("Node", ["address_info", "extra"], defaults=[None, tuple()])


class Property(Enum):
    EXTREME = auto()
    STABLE = auto()
    SYMMETRIC = auto()
    NON_SYMMETRIC = auto()
    HAS_GOSSIP = auto()
    ONE_TO_MANY = auto()
    MANY_TO_MANY = auto()



from typing import Dict, Any

from unsserv.common.utils import IConfig


class RWDConfig(IConfig):
    TIMEOUT = 2
    TTL = 10
    MAINTENANCE_SLEEP = 0.5
    QUANTUM = 0.1
    MORE_THAN_MAXIMUM = 30

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.TIMEOUT = config_dict.get("timeout", RWDConfig.TIMEOUT)
        self.TTL = config_dict.get("ttl", RWDConfig.TTL)
        self.MAINTENANCE_SLEEP = config_dict.get(
            "MAINTENANCE_SLEEP", RWDConfig.MAINTENANCE_SLEEP
        )
        self.QUANTUM = config_dict.get("quantum", RWDConfig.QUANTUM)
        self.MORE_THAN_MAXIMUM = config_dict.get(
            "more_than_maximum", RWDConfig.MORE_THAN_MAXIMUM
        )



from dataclasses import dataclass
from unsserv.common.structs import Node


@dataclass
class Sample:
    id: str
    origin_node: Node
    ttl: int


@dataclass
class SampleResult:
    sample_id: str
    result: Node



from enum import IntEnum, auto
from typing import Tuple, Sequence, List

from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.utils import parse_node
from unsserv.stable.searching.structs import Search, SearchResult, DataChange
from unsserv.stable.searching.typing import DataID

FIELD_COMMAND = "abloom-command"
FIELD_DATA_ID = "abloom-data-id"
FIELD_TTL = "abloom-ttl"
FIELD_ORIGIN_NODE = "abloom-origin-node"
FIELD_SEARCH_RESULT = "abloom-search-result"
FIELD_SEARCH_ID = "abloom-search-id"


class ABloomCommand(IntEnum):
    PUBLISH = auto()
    UNPUBLISH = auto()
    GET_FILTER = auto()
    SEARCH = auto()
    SEARCH_RESULT = auto()


class ABloomTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == ABloomCommand.PUBLISH:
            data_change: DataChange = data[0]
            message_data = {
                FIELD_COMMAND: ABloomCommand.PUBLISH,
                FIELD_DATA_ID: data_change.data_id,
                FIELD_TTL: data_change.ttl,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == ABloomCommand.UNPUBLISH:
            data_change: DataChange = data[0]  # type: ignore
            message_data = {
                FIELD_COMMAND: ABloomCommand.UNPUBLISH,
                FIELD_DATA_ID: data_change.data_id,
                FIELD_TTL: data_change.ttl,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == ABloomCommand.GET_FILTER:
            message_data = {FIELD_COMMAND: ABloomCommand.GET_FILTER}
            return Message(self.my_node, self.service_id, message_data)
        elif command == ABloomCommand.SEARCH:
            search: Search = data[0]
            message_data = {
                FIELD_COMMAND: ABloomCommand.SEARCH,
                FIELD_SEARCH_ID: search.id,
                FIELD_ORIGIN_NODE: search.origin_node,
                FIELD_TTL: search.ttl,
                FIELD_DATA_ID: search.data_id,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == ABloomCommand.SEARCH_RESULT:
            search_result: SearchResult = data[0]
            message_data = {
                FIELD_COMMAND: ABloomCommand.SEARCH_RESULT,
                FIELD_SEARCH_ID: search_result.search_id,
                FIELD_SEARCH_RESULT: search_result.result,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == ABloomCommand.PUBLISH:
            data_change = DataChange(
                data_id=message.data[FIELD_DATA_ID], ttl=message.data[FIELD_TTL]
            )
            return ABloomCommand.PUBLISH, [data_change]
        elif command == ABloomCommand.UNPUBLISH:
            data_change = DataChange(
                data_id=message.data[FIELD_DATA_ID], ttl=message.data[FIELD_TTL]
            )
            return ABloomCommand.UNPUBLISH, [data_change]
        elif command == ABloomCommand.GET_FILTER:
            return ABloomCommand.GET_FILTER, []
        elif command == ABloomCommand.SEARCH:
            search = Search(
                id=message.data[FIELD_SEARCH_ID],
                origin_node=parse_node(message.data[FIELD_ORIGIN_NODE]),
                ttl=message.data[FIELD_TTL],
                data_id=message.data[FIELD_DATA_ID],
            )
            return ABloomCommand.SEARCH, [search]
        elif command == ABloomCommand.SEARCH_RESULT:
            search_result = SearchResult(
                search_id=message.data[FIELD_SEARCH_ID],
                result=message.data[FIELD_SEARCH_RESULT],
            )
            return ABloomCommand.SEARCH_RESULT, [search_result]
        raise ValueError("Invalid Command")


class ABloomProtocol(AProtocol):
    def _get_new_transcoder(self):
        return ABloomTranscoder(self.my_node, self.service_id)

    async def publish(self, destination: Node, data_change: DataChange):
        message = self._transcoder.encode(ABloomCommand.PUBLISH, data_change)
        return await self._rpc.call_send_message(destination, message)

    async def unpublish(self, destination: Node, data_change: DataChange):
        message = self._transcoder.encode(ABloomCommand.UNPUBLISH, data_change)
        return await self._rpc.call_send_message(destination, message)

    async def get_filter(self, destination: Node):
        message = self._transcoder.encode(ABloomCommand.GET_FILTER)
        raw_filter = await self._rpc.call_send_message(destination, message)
        return parse_filter(raw_filter)

    async def search(self, destination: Node, search: Search):
        message = self._transcoder.encode(ABloomCommand.SEARCH, search)
        return await self._rpc.call_send_message(destination, message)

    async def search_result(self, destination: Node, search_result: SearchResult):
        message = self._transcoder.encode(ABloomCommand.SEARCH_RESULT, search_result)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_publish(self, handler: Handler):
        self._handlers[ABloomCommand.PUBLISH] = handler

    def set_handler_unpublish(self, handler: Handler):
        self._handlers[ABloomCommand.UNPUBLISH] = handler

    def set_handler_get_filter(self, handler: Handler):
        self._handlers[ABloomCommand.GET_FILTER] = handler

    def set_handler_search(self, handler: Handler):
        self._handlers[ABloomCommand.SEARCH] = handler

    def set_handler_search_result(self, handler: Handler):
        self._handlers[ABloomCommand.SEARCH_RESULT] = handler


def parse_filter(raw_filter: List[List[DataID]]):
    filter = []
    for depth in raw_filter:
        filter.append(set(depth))
    return filter



import asyncio
from functools import partial
from typing import Any, Dict

import pytest

from tests.utils import init_stable_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.stable.dissemination.plumtree.plumtree import Plumtree

init_stable_membership = init_stable_membership  # for flake8 compliance

DISSEMINATION_SERVICE_ID = "plumtree"

dissemination_data = b"data"
plum_events: Dict[Node, asyncio.Event] = {}


async def dissemination_handler(node: Node, data: Any):
    assert data == dissemination_data
    plum_events[node].set()


@pytest.mark.asyncio
@pytest.fixture
async def init_plum():
    plum = None
    r_plums = []

    async def _init_plum(hypa, r_hypas):
        nonlocal plum, r_plums
        plum = Plumtree(hypa)
        plum_events[hypa.my_node] = asyncio.Event()
        await plum.join(
            DISSEMINATION_SERVICE_ID,
            broadcast_handler=partial(dissemination_handler, plum.my_node),
        )
        for r_hypa in r_hypas:
            r_plum = Plumtree(r_hypa)
            plum_events[r_hypa.my_node] = asyncio.Event()
            await r_plum.join(
                DISSEMINATION_SERVICE_ID,
                broadcast_handler=partial(dissemination_handler, r_plum.my_node),
            )
            r_plums.append(r_plum)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return plum, r_plums

    try:
        yield _init_plum
    finally:
        await plum.leave()
        for r_plum in r_plums:
            await r_plum.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_stable_membership, init_plum, amount):
    hypa, r_hypas = await init_stable_membership(amount)
    plum, r_plums = await init_plum(hypa, r_hypas)


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_broadcast(init_stable_membership, init_plum, amount):
    hypa, r_hypas = await init_stable_membership(amount)
    plum, r_plums = await init_plum(hypa, r_hypas)

    await plum.broadcast(dissemination_data)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)

    plum_events_received = [plum_events[r_plum.my_node].is_set() for r_plum in r_plums]
    assert int(amount * 0.75) <= sum(plum_events_received)

    await plum.leave()
    for r_plum in r_plums:
        await r_plum.leave()



from math import ceil
from typing import Dict, Any

from unsserv.stable.membership.double_layered.config import DoubleLayeredConfig


class XBotConfig(DoubleLayeredConfig):
    UNBIASED_NODES = ceil(DoubleLayeredConfig.ACTIVE_VIEW_SIZE * 0.2)

    def load_from_dict(self, config_dict: Dict[str, Any]):
        super().load_from_dict(config_dict)
        self.UNBIASED_NODES = config_dict.get(
            "unbiased_nodes", XBotConfig.UNBIASED_NODES
        )



import asyncio
import random
from collections import OrderedDict
from typing import Any, List, Union

from unsserv.common.services_abc import IDisseminationService, IMembershipService
from unsserv.common.structs import Node, Property
from unsserv.common.typing import Handler
from unsserv.common.utils import get_random_id, HandlersManager
from unsserv.extreme.dissemination.lpbcast.config import LpbcastConfig
from unsserv.extreme.dissemination.lpbcast.protocol import LpbcastProtocol
from unsserv.extreme.dissemination.lpbcast.structs import Event
from unsserv.extreme.dissemination.lpbcast.typing import (
    EventId,
    EventData,
    EventOrigin,
)


class Lpbcast(IDisseminationService):
    properties = {Property.EXTREME, Property.MANY_TO_MANY}
    _protocol: LpbcastProtocol
    _handler_manager: HandlersManager
    _config: LpbcastConfig

    _events: "OrderedDict[EventId, List[Union[EventData, EventOrigin]]]"
    _events_digest: "OrderedDict[EventId, EventOrigin]"

    def __init__(self, membership: IMembershipService):
        self.my_node = membership.my_node
        self.membership = membership
        self._protocol = LpbcastProtocol(self.my_node)
        self._handler_manager = HandlersManager()
        self._config = LpbcastConfig()

        self._events = OrderedDict()
        self._events_digest = OrderedDict()

    async def join(self, service_id: str, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Dissemination")
        self.service_id = service_id
        self._handler_manager.add_handler(configuration["broadcast_handler"])
        await self._initialize_protocol()
        self._config.load_from_dict(configuration)
        self.running = True

    async def leave(self):
        if not self.running:
            return
        await self._protocol.stop()
        self._handler_manager.remove_all_handlers()
        self.running = False

    async def broadcast(self, data: bytes):
        if not self.running:
            raise RuntimeError("Dissemination service not running")
        assert isinstance(data, bytes)
        event_id = get_random_id()
        await self._handle_new_event(
            event_id, data, self.my_node, broadcast_origin=True
        )

    def add_broadcast_handler(self, handler: Handler):
        self._handler_manager.add_handler(handler)

    def remove_broadcast_handler(self, handler: Handler):
        self._handler_manager.remove_handler(handler)

    async def _handle_new_event(
        self,
        event_id: EventId,
        event_data: EventData,
        event_origin: Node,
        broadcast_origin=False,
    ):
        if event_id in self._events_digest:
            return
        self._events[event_id] = [event_data, event_origin]
        self._events_digest[event_id] = event_origin
        self._purge_events_threshold()
        asyncio.create_task(self._disseminate(event_id, event_data, event_origin))
        if not broadcast_origin:
            self._handler_manager.call_handlers(event_data)

    async def _disseminate(
        self, event_id: EventId, event_data: EventData, event_origin: Node
    ):
        candidate_neighbours = self.membership.get_neighbours()
        assert isinstance(candidate_neighbours, list)
        fanout = min(self._config.FANOUT, len(candidate_neighbours))
        for neighbour in random.choices(candidate_neighbours, k=fanout):
            try:
                digest = list(map(lambda e: (e[0], e[1]), self._events_digest.items()))
                event = Event(
                    id=event_id, data=event_data, origin=event_origin, digest=digest
                )
                await self._protocol.push_event(neighbour, event)
            except Exception:
                pass  # todo: log the error?

    def _purge_events_threshold(self):
        while self._config.THRESHOLD < len(self._events):
            self._events.popitem(last=False)
        while self._config.THRESHOLD < len(self._events_digest):
            self._events_digest.popitem(last=False)

    async def _retrieve_event(
        self, event_source: Node, event_id: EventId, event_origin: EventOrigin,
    ):

        try:
            event_data, _ = await self._protocol.retrieve_event(event_source, event_id)
            assert isinstance(event_data, bytes)
            return await self._handle_new_event(event_id, event_data, event_origin)
        except Exception:
            pass
        try:
            candidate_neighbours = self.membership.get_neighbours()
            assert isinstance(candidate_neighbours, list)
            random_neighbour = random.choice(candidate_neighbours)
            event_data, _ = await self._protocol.retrieve_event(
                random_neighbour, event_id
            )
            assert isinstance(event_data, bytes)
            return await self._handle_new_event(event_id, event_data, event_origin)
        except Exception:
            pass
        try:
            event_data, _ = await self._protocol.retrieve_event(event_origin, event_id)
            assert isinstance(event_data, bytes)
            return await self._handle_new_event(event_id, event_data, event_origin)
        except Exception:
            pass

    async def _handler_push_event(self, sender: Node, event: Event):
        asyncio.create_task(self._handle_new_event(event.id, event.data, event.origin,))
        for message_id, message_origin in event.digest:
            if message_id not in self._events_digest:  # not the one received
                asyncio.create_task(
                    self._retrieve_event(sender, message_id, message_origin)
                )

    async def _handler_retrieve_event(self, sender: Node, event_id: str):
        return self._events.get(event_id, None)

    async def _initialize_protocol(self):
        self._protocol.set_handler_push_event(self._handler_push_event)
        self._protocol.set_handler_retrieve_event(self._handler_retrieve_event)
        await self._protocol.start(self.service_id)



import asyncio
from functools import partial
from typing import Any, Dict

import pytest

from tests.utils import init_stable_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.stable.dissemination.brisa.brisa import Brisa

init_stable_membership = init_stable_membership  # for flake8 compliance

DISSEMINATION_SERVICE_ID = "brisa"

dissemination_data = b"data"
brisa_events: Dict[Node, asyncio.Event] = {}


async def dissemination_handler(node: Node, data: Any):
    assert data == dissemination_data
    brisa_events[node].set()


@pytest.mark.asyncio
@pytest.fixture
async def init_brisa():
    brisa = None
    r_brisas = []

    async def _init_brisa(hypa, r_hypas):
        nonlocal brisa, r_brisas
        brisa = Brisa(hypa)
        brisa_events[hypa.my_node] = asyncio.Event()
        await brisa.join(
            DISSEMINATION_SERVICE_ID,
            broadcast_handler=partial(dissemination_handler, brisa.my_node),
            im_root=True,
        )
        for r_hypa in r_hypas:
            r_brisa = Brisa(r_hypa)
            brisa_events[r_hypa.my_node] = asyncio.Event()
            await r_brisa.join(
                DISSEMINATION_SERVICE_ID,
                broadcast_handler=partial(dissemination_handler, r_brisa.my_node),
            )
            r_brisas.append(r_brisa)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return brisa, r_brisas

    try:
        yield _init_brisa
    finally:
        await brisa.leave()
        for r_brisa in r_brisas:
            await r_brisa.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_stable_membership, init_brisa, amount):
    hypa, r_hypas = await init_stable_membership(amount)
    brisa, r_brisas = await init_brisa(hypa, r_hypas)


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_broadcast(init_stable_membership, init_brisa, amount):
    hypa, r_hypas = await init_stable_membership(amount)
    brisa, r_brisas = await init_brisa(hypa, r_hypas)

    await brisa.broadcast(dissemination_data)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)

    brisa_events_received = [
        brisa_events[r_brisa.my_node].is_set() for r_brisa in r_brisas
    ]
    assert int(amount * 0.75) <= sum(brisa_events_received)

    await brisa.leave()
    for r_brisa in r_brisas:
        await r_brisa.leave()



from enum import IntEnum, auto
from typing import Tuple, Sequence

from unsserv.common.utils import parse_node
from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.extreme.searching.structs import WalkResult, Walk

FIELD_COMMAND = "kwalker-command"
FIELD_WALK_ID = "kwalker-walk-id"
FIELD_ORIGIN_NODE = "kwalker-origin-node"
FIELD_TTL = "kwalker-ttl"
FIELD_DATA_ID = "kwalker-data-id"
FIELD_WALK_RESULT = "kwalker-walk-result"


class KWalkerCommand(IntEnum):
    WALK = auto()
    WALK_RESULT = auto()


class KWalkerTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == KWalkerCommand.WALK:
            walk: Walk = data[0]
            message_data = {
                FIELD_COMMAND: KWalkerCommand.WALK,
                FIELD_WALK_ID: walk.id,
                FIELD_DATA_ID: walk.data_id,
                FIELD_ORIGIN_NODE: walk.origin_node,
                FIELD_TTL: walk.ttl,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == KWalkerCommand.WALK_RESULT:
            walk_result: WalkResult = data[0]
            message_data = {
                FIELD_COMMAND: KWalkerCommand.WALK_RESULT,
                FIELD_WALK_ID: walk_result.walk_id,
                FIELD_WALK_RESULT: walk_result.result,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == KWalkerCommand.WALK:
            walk = Walk(
                id=message.data[FIELD_WALK_ID],
                data_id=message.data[FIELD_DATA_ID],
                origin_node=parse_node(message.data[FIELD_ORIGIN_NODE]),
                ttl=message.data[FIELD_TTL],
            )
            return KWalkerCommand.WALK, [walk]
        elif command == KWalkerCommand.WALK_RESULT:
            walk_result = WalkResult(
                walk_id=message.data[FIELD_WALK_ID],
                result=message.data[FIELD_WALK_RESULT],
            )
            return KWalkerCommand.WALK_RESULT, [walk_result]
        raise ValueError("Invalid Command")


class KWalkerProtocol(AProtocol):
    def _get_new_transcoder(self):
        return KWalkerTranscoder(self.my_node, self.service_id)

    async def walk(self, destination: Node, walk: Walk):
        message = self._transcoder.encode(KWalkerCommand.WALK, walk)
        return await self._rpc.call_send_message(destination, message)

    async def walk_result(self, destination: Node, walk_result: WalkResult):
        message = self._transcoder.encode(KWalkerCommand.WALK_RESULT, walk_result)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_walk(self, handler: Handler):
        self._handlers[KWalkerCommand.WALK] = handler

    def set_handler_walk_result(self, handler: Handler):
        self._handlers[KWalkerCommand.WALK_RESULT] = handler



from statistics import mean
from typing import Any, Callable, Dict, Tuple

from unsserv.common.aggregation.config import AggregateType, AntiConfig
from unsserv.common.gossip.gossip import Gossip, IGossipSubscriber
from unsserv.common.gossip.typing import Payload
from unsserv.common.services_abc import IAggregationService, IMembershipService
from unsserv.common.structs import Property
from unsserv.common.typing import Handler
from unsserv.common.utils import HandlersManager

aggregate_functions: Dict[AggregateType, Callable] = {
    AggregateType.MEAN: mean,
    AggregateType.MAX: max,
    AggregateType.MIN: min,
}


class AntiEntropy(IAggregationService, IGossipSubscriber):
    """Aggregation Anti-Entropy service."""

    properties = {Property.EXTREME, Property.STABLE, Property.HAS_GOSSIP}
    gossip: Gossip
    _config: AntiConfig
    _handler_manager: HandlersManager
    _aggregate_value: Any

    def __init__(self, membership: IMembershipService):
        self.my_node = membership.my_node
        if Property.HAS_GOSSIP not in membership.properties:
            raise ValueError(
                "Invalid membership service. "
                "Membership must contain a 'gossip' attribute"
            )
        self.membership = membership
        self.gossip = getattr(membership, "gossip")
        self._aggregate_value = None
        self._handler_manager = HandlersManager()
        self._config = AntiConfig()

    async def join(self, service_id: str, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Aggregation")
        self._config.load_from_dict(configuration)
        self._aggregate_value = configuration["aggregate_value"]
        self.service_id = service_id
        self.gossip.subscribe(self)
        self.running = True

    async def leave(self):
        if not self.running:
            return
        self.gossip.unsubscribe(self)
        self._handler_manager.remove_all_handlers()
        self._aggregate_value = None
        self.running = False

    async def get_aggregate(self) -> Any:
        if not self.running:
            raise RuntimeError("Aggregation service not running")
        return self._aggregate_value

    def add_aggregate_handler(self, handler: Handler):
        self._handler_manager.add_handler(handler)

    def remove_aggregate_handler(self, handler: Handler):
        self._handler_manager.remove_handler(handler)

    async def receive_payload(self, payload: Payload):
        """IGossipSubscriber implementation."""
        aggregate_function = aggregate_functions[self._config.AGGREGATE_TYPE]
        assert callable(aggregate_function)
        neighbor_aggregate = payload.get(self.service_id, None)
        if not neighbor_aggregate:
            return
        self._aggregate_value = aggregate_function(
            [self._aggregate_value, neighbor_aggregate]
        )
        self._handler_manager.call_handlers(self._aggregate_value)

    async def get_payload(self) -> Tuple[Any, Any]:
        """IGossipSubscriber implementation."""
        return self.service_id, self._aggregate_value



import asyncio
from collections import Counter
from math import ceil

import pytest

from tests.utils import get_random_nodes
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.extreme.membership import newscast

node = Node(("127.0.0.1", 7771))
MEMBERSHIP_SERVICE_ID = "newscast"


@pytest.mark.asyncio
@pytest.fixture
async def init_newscast():
    newc = None
    r_newcs = []

    async def _init_newscast(amount):
        nonlocal newc, r_newcs
        newc = newscast.Newscast(node)
        await newc.join(MEMBERSHIP_SERVICE_ID)
        r_nodes = get_random_nodes(amount)
        for i, r_node in enumerate(r_nodes):
            r_newc = newscast.Newscast(r_node)
            await r_newc.join(
                MEMBERSHIP_SERVICE_ID, bootstrap_nodes=[node] + r_nodes[:i]
            )
            r_newcs.append(r_newc)
        return newc, r_newcs, r_nodes

    try:
        yield _init_newscast
    finally:
        await newc.leave()
        for r_newc in r_newcs:
            await r_newc.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_newscast_join(init_newscast, amount):
    newc, r_newcs, r_nodes = await init_newscast(amount)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)

    all_nodes = set(
        [
            item
            for sublist in map(lambda n: n.get_neighbours(), r_newcs + [newc])
            for item in sublist
        ]
    )
    assert amount * 0.9 < len(all_nodes)

    neighbours = newc.get_neighbours()
    assert min(amount, GossipConfig.LOCAL_VIEW_SIZE) <= len(neighbours)
    for neighbour in neighbours:
        assert neighbour in r_nodes

    for r_newc in r_newcs:
        r_neighbours = r_newc.get_neighbours()
        assert min(amount, GossipConfig.LOCAL_VIEW_SIZE) <= len(r_neighbours)
        for r_neighbour in r_neighbours:
            assert r_neighbour in r_nodes or r_neighbour == node


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        GossipConfig.LOCAL_VIEW_SIZE + 1,
        GossipConfig.LOCAL_VIEW_SIZE + 5,
        GossipConfig.LOCAL_VIEW_SIZE + 100,
    ],
)
async def test_newscast_leave(init_newscast, amount):
    newc, r_newcs, r_nodes = await init_newscast(amount)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)

    await newc.leave()
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 40)

    all_nodes = Counter(
        [
            item
            for sublist in map(lambda n: n.get_neighbours(), r_newcs)
            for item in sublist
        ]
    )
    nodes_ten_percent = ceil(amount * 0.2)
    assert node not in all_nodes.keys() or node in set(
        map(lambda p: p[0], all_nodes.most_common()[-nodes_ten_percent:])
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount",
    [
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 1,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 5,
        (GossipConfig.LOCAL_VIEW_SIZE * 2) + 100,
    ],
)
async def test_newscast_handler(init_newscast, amount):
    newc, r_newcs, r_nodes = await init_newscast(amount)

    handler_event = asyncio.Event()

    async def handler(neighbours):
        nonlocal handler_event
        handler_event.set()

    newc.add_neighbours_handler(handler)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
    assert handler_event.is_set()



from enum import IntEnum, auto
from typing import Tuple, Sequence

from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.utils import parse_node
from unsserv.stable.clustering.structs import Replace

FIELD_COMMAND = "xbot-command"
FIELD_OLD_NODE = "xbot-old-node"
FIELD_ORIGIN_NODE = "xbot-origin-node"


class XBotCommand(IntEnum):
    OPTIMIZATION = auto()
    REPLACE = auto()
    SWITCH = auto()


class XBotTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == XBotCommand.OPTIMIZATION:
            old_node: Node = data[0]
            message_data = {
                FIELD_COMMAND: XBotCommand.OPTIMIZATION,
                FIELD_OLD_NODE: old_node,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == XBotCommand.REPLACE:
            replace: Replace = data[0]
            message_data = {
                FIELD_ORIGIN_NODE: replace.origin_node,
                FIELD_OLD_NODE: replace.old_node,
                FIELD_COMMAND: XBotCommand.REPLACE,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == XBotCommand.SWITCH:
            origin_node: Node = data[0]
            message_data = {
                FIELD_COMMAND: XBotCommand.SWITCH,
                FIELD_ORIGIN_NODE: origin_node,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == XBotCommand.OPTIMIZATION:
            old_node: Node = parse_node(message.data[FIELD_OLD_NODE])
            return XBotCommand.OPTIMIZATION, [old_node]
        elif command == XBotCommand.REPLACE:
            replace = Replace(
                old_node=parse_node(message.data[FIELD_OLD_NODE]),
                origin_node=parse_node(message.data[FIELD_ORIGIN_NODE]),
            )
            return XBotCommand.REPLACE, [replace]
        elif command == XBotCommand.SWITCH:
            origin_node: Node = parse_node(message.data[FIELD_ORIGIN_NODE])
            return XBotCommand.SWITCH, [origin_node]
        raise ValueError("Invalid Command")


class XBotProtocol(AProtocol):
    def _get_new_transcoder(self):
        return XBotTranscoder(self.my_node, self.service_id)

    async def optimization(self, destination: Node, old_node: Node):
        message = self._transcoder.encode(XBotCommand.OPTIMIZATION, old_node)
        return await self._rpc.call_send_message(destination, message)

    async def replace(self, destination: Node, replace: Replace):
        message = self._transcoder.encode(XBotCommand.REPLACE, replace)
        return await self._rpc.call_send_message(destination, message)

    async def switch(self, destination: Node, origin_node: Node) -> int:
        message = self._transcoder.encode(XBotCommand.SWITCH, origin_node)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_optimization(self, handler: Handler):
        self._handlers[XBotCommand.OPTIMIZATION] = handler

    def set_handler_replace(self, handler: Handler):
        self._handlers[XBotCommand.REPLACE] = handler

    def set_handler_switch(self, handler: Handler):
        self._handlers[XBotCommand.SWITCH] = handler



import asyncio
import random
from typing import Any, List, Dict

from unsserv.common.errors import ServiceError
from unsserv.common.services_abc import IDisseminationService, IMembershipService
from unsserv.common.structs import Node, Property
from unsserv.common.typing import Handler
from unsserv.common.utils import get_random_id, HandlersManager, stop_task
from unsserv.extreme.dissemination.mon.config import MonConfig
from unsserv.extreme.dissemination.mon.protocol import MonProtocol
from unsserv.extreme.dissemination.mon.structs import Session, Broadcast
from unsserv.extreme.dissemination.mon.typing import BroadcastID


class Mon(IDisseminationService):
    properties = {Property.EXTREME, Property.ONE_TO_MANY}
    _protocol: MonProtocol
    _handler_manager: HandlersManager
    _config: MonConfig

    _levels: Dict[BroadcastID, int]
    _children: Dict[BroadcastID, List[Node]]
    _parents: Dict[BroadcastID, List[Node]]
    _received_data: Dict[
        BroadcastID, Any
    ]  # stores the data received from each broadcast (for avoiding duplicates)
    _children_ready_events: Dict[BroadcastID, asyncio.Event]
    _cleanup_tasks: List[asyncio.Task]

    def __init__(self, membership: IMembershipService):
        self.my_node = membership.my_node
        self.membership = membership
        self._protocol = MonProtocol(self.my_node)
        self._handler_manager = HandlersManager()
        self._config = MonConfig()

        self._children = {}
        self._parents = {}
        self._levels = {}
        self._received_data = {}

        self._children_ready_events = {}
        self._cleanup_tasks = []

    async def join(self, service_id: str, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Dissemination")
        self.service_id = service_id
        await self._initialize_protocol()
        self._handler_manager.add_handler(configuration["broadcast_handler"])
        self._config.load_from_dict(configuration)
        self.running = True

    async def leave(self):
        if not self.running:
            return
        for task in self._cleanup_tasks:
            await stop_task(task)
        self._handler_manager.remove_all_handlers()
        await self._protocol.stop()
        self.running = False

    async def broadcast(self, data: bytes):
        if not self.running:
            raise RuntimeError("Dissemination service not running")
        assert isinstance(data, bytes)
        broadcast_id = await asyncio.wait_for(
            self._build_dag(), timeout=self._config.TIMEOUT
        )
        await asyncio.wait_for(
            self._disseminate(broadcast_id, data), timeout=self._config.TIMEOUT
        )

    def add_broadcast_handler(self, handler: Handler):
        self._handler_manager.add_handler(handler)

    def remove_broadcast_handler(self, handler: Handler):
        self._handler_manager.remove_handler(handler)

    async def _build_dag(self) -> str:
        broadcast_id = get_random_id()
        self._levels[broadcast_id] = 0
        self._children_ready_events[broadcast_id] = asyncio.Event()
        candidate_children = self.membership.get_neighbours()
        asyncio.create_task(
            self._initialize_children(
                broadcast_id, candidate_children, broadcast_origin=True
            )
        )
        return broadcast_id

    async def _initialize_children(
        self, broadcast_id: str, neighbours: list, broadcast_origin=False
    ):
        children: List[Node] = []
        fanout = min(len(neighbours), self._config.FANOUT)
        session = Session(broadcast_id=broadcast_id, level=self._levels[broadcast_id])
        while neighbours and len(children) <= fanout:
            child = random.choice(neighbours)
            neighbours.remove(child)
            try:
                session_ok = await self._protocol.session(child, session)
            except ConnectionError:
                continue
            if session_ok:
                children.append(child)
        if broadcast_origin and len(children) == 0:
            raise ServiceError("Unable to peer with neighbours for disseminating")
        self._children[broadcast_id] = children
        self._children_ready_events[broadcast_id].set()

        self._cleanup_tasks.append(
            asyncio.create_task(self._cleanup_tree(broadcast_id))
        )

    async def _disseminate(self, broadcast_id: str, data: Any):
        await self._children_ready_events[
            broadcast_id
        ].wait()  # wait children to initialize
        broadcast = Broadcast(id=broadcast_id, data=data)
        pushed_amount = 0
        for child in self._children[broadcast_id]:
            try:
                await self._protocol.push(child, broadcast)
                pushed_amount += 1
            except ConnectionError:
                pass
        if pushed_amount == 0:
            raise ServiceError("Unable to peer with neighbours for disseminating")

    async def _cleanup_tree(self, broadcast_id: str):
        await asyncio.sleep(self._config.TREE_LIFE_TIME)
        if broadcast_id in self._levels:
            del self._levels[broadcast_id]
        if broadcast_id in self._children_ready_events:
            del self._children_ready_events[broadcast_id]
        if broadcast_id in self._children:
            del self._children[broadcast_id]
        if broadcast_id in self._received_data:
            del self._received_data[broadcast_id]

    async def _handler_session(self, sender: Node, session: Session):
        first_time = session.broadcast_id not in self._levels
        if first_time:
            self._parents[session.broadcast_id] = [sender]
            self._levels[session.broadcast_id] = session.level + 1
            candidate_children = list(set(self.membership.get_neighbours()) - {sender})
            self._children_ready_events[session.broadcast_id] = asyncio.Event()
            asyncio.create_task(
                self._initialize_children(session.broadcast_id, candidate_children)
            )
        else:
            if self._levels[session.broadcast_id] <= session.level:
                return False
            if sender not in self._parents[session.broadcast_id]:
                self._parents[session.broadcast_id].append(sender)
        return True

    async def _handler_push(self, sender: Node, broadcast: Broadcast):
        # if already received data, ignores it
        if (
            broadcast.id in self._children_ready_events
            and broadcast.id not in self._received_data
        ):
            self._received_data[broadcast.id] = broadcast.data
            asyncio.create_task(self._disseminate(broadcast.id, broadcast.data))
            self._handler_manager.call_handlers(broadcast.data)

    async def _initialize_protocol(self):
        self._protocol.set_handler_session(self._handler_session)
        self._protocol.set_handler_push(self._handler_push)
        await self._protocol.start(self.service_id)



import asyncio
from collections import Counter
from typing import Any, List, Tuple

import pytest

import unsserv.common.gossip.config
import unsserv.common.gossip.gossip
import unsserv.common.gossip.structs
from tests.utils import get_random_nodes
from unsserv.common.gossip import gossip
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.gossip.gossip import View
from unsserv.common.gossip.typing import Payload
from unsserv.common.structs import Node
from unsserv.common.utils import parse_node

node = Node(("127.0.0.1", 7771))
SERVICE_ID = "gossip"


def test_view_selection():
    r_nodes = get_random_nodes(GossipConfig.LOCAL_VIEW_SIZE * 2, first_port=7772)
    view = Counter(dict(map(lambda n: (n[1], n[0] + 1), enumerate(r_nodes))))
    gsp = gossip.Gossip(node, SERVICE_ID)

    gsp._config.VIEW_SELECTION = unsserv.common.gossip.config.ViewSelectionPolicy.HEAD
    assert set(r_nodes[: GossipConfig.LOCAL_VIEW_SIZE]) == set(
        gsp._select_view(view).keys()
    )

    gsp._config.VIEW_SELECTION = unsserv.common.gossip.config.ViewSelectionPolicy.TAIL
    assert set(r_nodes[GossipConfig.LOCAL_VIEW_SIZE :]) == set(
        gsp._select_view(view).keys()
    )

    def ranking(node: Node):
        return node.address_info[1] % 2

    def selection_ranking(view: View) -> List[Node]:
        return sorted(view.keys(), key=ranking)

    gsp._config.VIEW_SELECTION = unsserv.common.gossip.config.ViewSelectionPolicy.HEAD
    gsp.custom_selection_ranking = selection_ranking
    assert set(sorted(r_nodes, key=ranking)[: GossipConfig.LOCAL_VIEW_SIZE]) == set(
        gsp._select_view(view).keys()
    )


def test_peer_selection():
    r_nodes = get_random_nodes(GossipConfig.LOCAL_VIEW_SIZE, first_port=7772)
    view = Counter(dict(map(lambda n: (n[1], n[0] + 1), enumerate(r_nodes))))
    gsp = gossip.Gossip(node, SERVICE_ID)

    gsp._config.PEER_SELECTION = unsserv.common.gossip.config.PeerSelectionPolicy.HEAD
    assert r_nodes[0] == gsp._select_peer(view)

    gsp._config.PEER_SELECTION = unsserv.common.gossip.config.PeerSelectionPolicy.TAIL
    assert r_nodes[-1] == gsp._select_peer(view)


def test_increase_hop_count():
    r_nodes = get_random_nodes(GossipConfig.LOCAL_VIEW_SIZE, first_port=7772)
    view = Counter(dict(map(lambda n: (n[1], n[0] + 1), enumerate(r_nodes))))
    gsp = gossip.Gossip(node, SERVICE_ID)

    assert view + Counter(view.keys()) == gsp._increase_hop_count(view)


def test_merge():
    r_nodes = get_random_nodes(GossipConfig.LOCAL_VIEW_SIZE, first_port=7772)
    view1 = Counter(dict(map(lambda n: (n[1], n[0] + 1), enumerate(r_nodes))))
    gsp = gossip.Gossip(node, SERVICE_ID)

    view2 = view1.copy()
    view2[r_nodes[0]] = 99

    assert view1 != view2
    assert view1 == gsp._merge_views(view1, view2)


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_gossiping(amount):
    r_nodes = get_random_nodes(amount, first_port=7772)
    gsp = gossip.Gossip(node, SERVICE_ID)
    await gsp.start()

    r_gsps = []
    for i, r_node in enumerate(r_nodes):
        r_gsp = gossip.Gossip(r_node, SERVICE_ID, [node] + r_nodes[:i])
        await r_gsp.start()
        r_gsps.append(r_gsp)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)

    for r_gsp in r_gsps:
        assert min(amount, GossipConfig.LOCAL_VIEW_SIZE) <= len(r_gsp.local_view)

    await gsp.stop()
    for r_gsp in r_gsps:
        await r_gsp.stop()


@pytest.mark.asyncio
async def test_subscriber():
    class Subscriber(unsserv.common.gossip.gossip.IGossipSubscriber):
        service_id = "sub"

        def __init__(self, my_node, expected_node):
            self.my_node = my_node
            self.expected_node = expected_node
            self.is_payload_received = False

        async def receive_payload(self, payload: Payload):
            node = parse_node(payload[Subscriber.service_id])
            assert node == self.expected_node
            self.is_payload_received = True

        async def get_payload(self) -> Tuple[Any, Any]:
            return Subscriber.service_id, self.my_node

    r_node = get_random_nodes(1, first_port=7772)[0]
    gsp = gossip.Gossip(node, SERVICE_ID)
    sub_gsp = Subscriber(node, r_node)
    gsp.subscribe(sub_gsp)

    r_gsp = gossip.Gossip(r_node, SERVICE_ID, [node])
    sub_r_gsp = Subscriber(r_node, node)
    r_gsp.subscribe(sub_r_gsp)

    await gsp.start()
    await r_gsp.start()

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 3)

    assert sub_gsp.is_payload_received
    assert sub_r_gsp.is_payload_received

    await gsp.stop()
    await r_gsp.stop()



from enum import IntEnum, auto
from typing import Tuple, Sequence

from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.stable.dissemination.brisa.typing import BroadcastLevel


FIELD_COMMAND = "brisa-command"
FIELD_LEVEL = "brisa-level"
FIELD_BROADCAST_DATA = "brisa-broadcast-data"


class BrisaCommand(IntEnum):
    SESSION = auto()
    PUSH = auto()
    IM_YOUR_CHILD = auto()
    BECOME_MY_PARENT = auto()


class BrisaTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == BrisaCommand.SESSION:
            level: BroadcastLevel = data[0]
            message_data = {
                FIELD_COMMAND: BrisaCommand.SESSION,
                FIELD_LEVEL: level,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == BrisaCommand.PUSH:
            broadcast_data: bytes = data[0]
            message_data = {
                FIELD_COMMAND: BrisaCommand.PUSH,
                FIELD_BROADCAST_DATA: broadcast_data,  # type:ignore
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == BrisaCommand.IM_YOUR_CHILD:
            message_data = {
                FIELD_COMMAND: BrisaCommand.IM_YOUR_CHILD,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == BrisaCommand.BECOME_MY_PARENT:
            message_data = {
                FIELD_COMMAND: BrisaCommand.BECOME_MY_PARENT,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == BrisaCommand.SESSION:
            level: BroadcastLevel = message.data[FIELD_LEVEL]
            return BrisaCommand.SESSION, [level]
        elif command == BrisaCommand.PUSH:
            data: bytes = message.data[FIELD_BROADCAST_DATA]
            return BrisaCommand.PUSH, [data]
        elif command == BrisaCommand.IM_YOUR_CHILD:
            return BrisaCommand.IM_YOUR_CHILD, []
        elif command == BrisaCommand.BECOME_MY_PARENT:
            return BrisaCommand.BECOME_MY_PARENT, []
        raise ValueError("Invalid Command")


class BrisaProtocol(AProtocol):
    def _get_new_transcoder(self):
        return BrisaTranscoder(self.my_node, self.service_id)

    async def session(self, destination: Node, level: BroadcastLevel) -> bool:
        message = self._transcoder.encode(BrisaCommand.SESSION, level)
        return await self._rpc.call_send_message(destination, message)

    async def push(self, destination: Node, data: bytes):
        message = self._transcoder.encode(BrisaCommand.PUSH, data)
        return await self._rpc.call_send_message(destination, message)

    async def im_your_child(self, destination: Node) -> bool:
        message = self._transcoder.encode(BrisaCommand.IM_YOUR_CHILD)
        return await self._rpc.call_send_message(destination, message)

    async def become_my_parent(self, destination: Node) -> BroadcastLevel:
        message = self._transcoder.encode(BrisaCommand.BECOME_MY_PARENT)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_session(self, handler: Handler):
        self._handlers[BrisaCommand.SESSION] = handler

    def set_handler_push(self, handler: Handler):
        self._handlers[BrisaCommand.PUSH] = handler

    def set_handler_im_your_child(self, handler: Handler):
        self._handlers[BrisaCommand.IM_YOUR_CHILD] = handler

    def set_handler_become_my_parent(self, handler: Handler):
        self._handlers[BrisaCommand.BECOME_MY_PARENT] = handler



import asyncio
import random
from typing import Any, Callable, List

from unsserv.common.services_abc import IMembershipService, IClusteringService
from unsserv.common.structs import Node, Property
from unsserv.common.typing import Handler
from unsserv.common.utils import stop_task
from unsserv.stable.clustering.config import XBotConfig
from unsserv.stable.clustering.protocol import XBotProtocol
from unsserv.stable.clustering.structs import Replace
from unsserv.stable.membership.double_layered.double_layered import IDoubleLayered

RankingFunction = Callable[[Node], Any]


class XBot(IClusteringService, IDoubleLayered):
    properties = {Property.STABLE, Property.SYMMETRIC}
    _protocol: XBotProtocol
    _config: XBotConfig

    _local_view_optimize_task: asyncio.Task

    def __init__(self, membership: IMembershipService):
        super().__init__(membership.my_node)
        self.membership = membership
        self._protocol = XBotProtocol(membership.my_node)
        self._config = XBotConfig()

        self._ranking_function: RankingFunction

    async def join(self, service_id: Any, **configuration: Any):
        if self.running:
            raise RuntimeError("Already running Clustering")
        self.service_id = service_id
        await self._initialize_protocol()
        self._config.load_from_dict(configuration)
        self._ranking_function = configuration["ranking_function"]
        await self._start_two_layered(f"double_layered-{service_id}")
        self._local_view_optimize_task = asyncio.create_task(
            self._optimize_active_view_loop()
        )
        self.running = True

    async def leave(self):
        if not self.running:
            return
        if self._local_view_optimize_task:
            await stop_task(self._local_view_optimize_task)
        await self._protocol.stop()
        await self._stop_two_layered()
        self.running = False

    def get_neighbours(self) -> List[Node]:
        return list(self._active_view)

    def add_neighbours_handler(self, handler: Handler):
        if not self.running:
            raise RuntimeError("Membership service not running")
        self._handler_manager.add_handler(handler)

    def remove_neighbours_handler(self, handler: Handler):
        self._handler_manager.remove_handler(handler)

    def _get_passive_view_nodes(self):
        return self.membership.get_neighbours()

    async def _optimize_active_view_loop(self):
        await asyncio.sleep(self._config.MAINTENANCE_SLEEP)
        while True:
            old_active_view = self._active_view.copy()
            await asyncio.sleep(self._config.MAINTENANCE_SLEEP)
            if len(self._active_view) >= self._config.ACTIVE_VIEW_SIZE:
                await self._optimize_active_view()  # todo: create task instead?
            self._call_handler_if_view_changed(old_active_view)

    async def _optimize_active_view(self):
        candidate_neighbours = self.membership.get_neighbours()
        if not candidate_neighbours:
            return
        candidate_neighbours = random.sample(
            candidate_neighbours,
            min(len(candidate_neighbours), self._config.ACTIVE_VIEW_SIZE),
        )
        biasable_nodes = list(sorted(self._active_view, key=self._ranking_function))[
            self._config.UNBIASED_NODES :
        ]
        for old_node in biasable_nodes:
            if not candidate_neighbours:
                return None
            candidate_node = candidate_neighbours.pop()
            if self._get_the_best(candidate_node, old_node) == candidate_node:
                asyncio.create_task(self._optimize(old_node, candidate_node))

    async def _optimize(self, old_node: Node, new_node: Node):
        is_optimized = await self._protocol.optimization(new_node, old_node)
        if not is_optimized:
            return
        if old_node in self._active_view:
            asyncio.create_task(self._doble_layered_protocol.disconnect(old_node))
            self._active_view.remove(old_node)
        self._active_view.add(new_node)

    def _get_the_best(self, node1: Node, node2: Node):
        return (
            node1
            if self._ranking_function(node1) < self._ranking_function(node2)
            else node2
        )

    async def _handler_optimization(self, sender: Node, old_node: Node):
        if len(self._active_view) < self._config.ACTIVE_VIEW_SIZE:
            self._active_view.add(sender)
            return True

        replace_node = list(sorted(self._active_view, key=self._ranking_function))[-1]
        replace = Replace(old_node=old_node, origin_node=sender)
        is_replaced = await self._protocol.replace(replace_node, replace)

        if is_replaced:
            self._active_view.remove(replace_node)
            self._active_view.add(sender)
            return True
        return False

    async def _handler_replace(self, sender: Node, replace: Replace):
        if self._get_the_best(replace.old_node, sender) == sender:
            return False
        else:
            is_switched = await self._protocol.switch(
                replace.old_node, replace.origin_node
            )
            if is_switched:
                if sender in self._active_view:
                    self._active_view.remove(sender)
                self._active_view.add(replace.old_node)
                return True
            return False

    async def _handler_switch(self, sender: Node, origin_node: Node):
        if origin_node in self._active_view:
            self._active_view.remove(origin_node)
            self._active_view.add(sender)
            return True
        return False

    async def _initialize_protocol(self):
        self._protocol.set_handler_optimization(self._handler_optimization)
        self._protocol.set_handler_replace(self._handler_replace)
        self._protocol.set_handler_switch(self._handler_switch)
        await self._protocol.start(self.service_id)



import asyncio
import random
from math import ceil
from typing import Dict

import pytest

from tests.utils import init_stable_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.common.utils import get_random_id
from unsserv.stable.searching.abloom import ABloom
from unsserv.stable.searching.config import ABloomConfig
from unsserv.stable.searching.structs import Search

init_stable_membership = init_stable_membership  # for flake8 compliance

SAMPLING_SERVICE_ID = "abloom"


@pytest.mark.asyncio
@pytest.fixture
async def init_abloom():
    abloom = None
    r_ablooms = []

    async def _init_abloom(xbot, r_xbots):
        nonlocal abloom, r_ablooms
        abloom = ABloom(xbot)
        await abloom.join(SAMPLING_SERVICE_ID, ttl=2)
        for r_xbot in r_xbots:
            r_abloom = ABloom(r_xbot)
            await r_abloom.join(SAMPLING_SERVICE_ID, ttl=2)
            r_ablooms.append(r_abloom)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return abloom, r_ablooms

    try:
        yield _init_abloom
    finally:
        await abloom.leave()
        for r_abloom in r_ablooms:
            await r_abloom.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_stable_membership, init_abloom, amount):
    xbot, r_xbots = await init_stable_membership(amount)
    abloom, r_ablooms = await init_abloom(xbot, r_xbots)

    await asyncio.sleep(5)
    assert xbot._active_view == set(abloom._neighbours)


def check_filter(node: Node, abloom_dict: Dict[Node, ABloom], search: Search):
    abloom, checked = abloom_dict[node]
    if search.origin_node == node or checked or search.ttl < 1:
        return

    assert (
        search.data_id
        in abloom._abloom_filters[search.origin_node][ABloomConfig.DEPTH - search.ttl]
    )
    abloom_dict[node][1] = True  # set as 'checked'
    next_search = Search(
        id=search.id,
        origin_node=search.origin_node,
        ttl=search.ttl - 1,
        data_id=search.data_id,
    )
    for neighbour in abloom._neighbours:
        check_filter(neighbour, abloom_dict, next_search)


@pytest.mark.asyncio
async def test_publish_unpublish(init_stable_membership, init_abloom):
    xbot, r_xbots = await init_stable_membership(1)
    abloom, r_ablooms = await init_abloom(xbot, r_xbots)

    data = b"test-data"
    data_id = get_random_id()
    await abloom.publish(data_id, data)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 5)

    ablooms_dict = {abl.my_node: [abl, False] for abl in r_ablooms + [abloom]}
    search = Search(
        id="id", origin_node=abloom.my_node, ttl=ABloomConfig.DEPTH, data_id=data_id
    )
    for neighbour in abloom._neighbours:
        check_filter(neighbour, ablooms_dict, search)

    with pytest.raises(KeyError):
        await abloom.publish(data_id, data)

    with pytest.raises(KeyError):
        await abloom.publish(data_id, data + b"more-data")

    with pytest.raises(KeyError):
        await abloom.unpublish(data_id + "wrong-id")

    await abloom.unpublish(data_id)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount,replication_percent",
    [
        (amount, repl_percent)
        for amount in [
            GossipConfig.LOCAL_VIEW_SIZE + 1,
            GossipConfig.LOCAL_VIEW_SIZE + 5,
            GossipConfig.LOCAL_VIEW_SIZE + 100,
        ]
        for repl_percent in [0.3, 0.6, 0.9]
    ],
)
async def test_search(init_stable_membership, init_abloom, amount, replication_percent):
    xbot, r_xbots = await init_stable_membership(amount)
    abloom, r_ablooms = await init_abloom(xbot, r_xbots)
    data_id = get_random_id()
    data = b"data"
    replication_amount = ceil(amount * replication_percent)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 10)
    for r_abloom in random.sample(r_ablooms, replication_amount):
        await r_abloom.publish(data_id, data)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 5)

    found_data = await abloom.search(data_id)
    assert found_data


@pytest.mark.asyncio
async def test_search_fail(init_stable_membership, init_abloom):
    xbot, r_xbots = await init_stable_membership(100)
    abloom, r_ablooms = await init_abloom(xbot, r_xbots)
    data_id = get_random_id()
    result = await abloom.search(data_id)
    assert not result



from collections import Counter
from enum import IntEnum, auto
from typing import Tuple, Sequence, Dict, Any

from unsserv.common.gossip.structs import PushData
from unsserv.common.gossip.typing import Payload
from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.typing import View

FIELD_COMMAND = "gossip-command"
FIELD_VIEW = "gossip-view"
FIELD_PAYLOAD = "gossip-payload"


class GossipCommand(IntEnum):
    PUSH = auto()
    PULL = auto()
    PUSHPULL = auto()


class GossipTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == GossipCommand.PUSH:
            push_data: PushData = data[0]
            message_data = {
                FIELD_COMMAND: GossipCommand.PUSH,
                FIELD_VIEW: push_data.view,
                FIELD_PAYLOAD: push_data.payload,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == GossipCommand.PULL:
            payload: Payload = data[0]
            message_data = {FIELD_COMMAND: GossipCommand.PULL, FIELD_PAYLOAD: payload}
            return Message(self.my_node, self.service_id, message_data)
        elif command == GossipCommand.PUSHPULL:
            push_data: PushData = data[0]  # type: ignore
            message_data = {
                FIELD_COMMAND: GossipCommand.PUSHPULL,
                FIELD_VIEW: push_data.view,
                FIELD_PAYLOAD: push_data.payload,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == GossipCommand.PUSH:
            push_data = PushData(
                view=_parse_view(message.data[FIELD_VIEW]),
                payload=message.data[FIELD_PAYLOAD],
            )
            return GossipCommand.PUSH, [push_data]
        elif command == GossipCommand.PULL:
            payload = message.data[FIELD_PAYLOAD]
            return GossipCommand.PULL, [payload]
        elif command == GossipCommand.PUSHPULL:
            push_data = PushData(
                view=_parse_view(message.data[FIELD_VIEW]),
                payload=message.data[FIELD_PAYLOAD],
            )
            return GossipCommand.PUSHPULL, [push_data]
        raise ValueError("Invalid Command")


class GossipProtocol(AProtocol):
    def _get_new_transcoder(self):
        return GossipTranscoder(self.my_node, self.service_id)

    async def push(self, destination: Node, push_data: PushData):
        message = self._transcoder.encode(GossipCommand.PUSH, push_data)
        return await self._rpc.call_send_message(destination, message)

    async def pull(self, destination: Node, payload: Payload) -> PushData:
        message = self._transcoder.encode(GossipCommand.PULL, payload)
        encoded_push_data = await self._rpc.call_send_message(destination, message)
        return _parse_push_data(encoded_push_data)

    async def pushpull(self, destination: Node, push_data: PushData) -> PushData:
        message = self._transcoder.encode(GossipCommand.PUSHPULL, push_data)
        encoded_push_data = await self._rpc.call_send_message(destination, message)
        return _parse_push_data(encoded_push_data)

    def set_handler_push(self, handler: Handler):
        self._handlers[GossipCommand.PUSH] = handler

    def set_handler_pull(self, handler: Handler):
        self._handlers[GossipCommand.PULL] = handler

    def set_handler_pushpull(self, handler: Handler):
        self._handlers[GossipCommand.PUSHPULL] = handler


def _parse_view(raw_view: Dict[Any, Any]) -> View:
    return Counter(dict(map(lambda n: (Node(*n[0]), n[1]), raw_view.items())))


def _parse_push_data(raw_push_data: Dict[str, Any]) -> PushData:
    return PushData(
        view=_parse_view(raw_push_data["view"]), payload=raw_push_data["payload"]
    )



import asyncio
import random
from math import ceil

import pytest

from tests.utils import init_extreme_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.utils import get_random_id
from unsserv.extreme.searching.k_walker import KWalker

init_extreme_membership = init_extreme_membership  # for flake8 compliance

SAMPLING_SERVICE_ID = "kwalker"


@pytest.mark.asyncio
@pytest.fixture
async def init_kwalker():
    kwalker = None
    r_kwalkers = []

    async def _init_kwalker(newc, r_newcs):
        nonlocal kwalker, r_kwalkers
        kwalker = KWalker(newc)
        await kwalker.join(SAMPLING_SERVICE_ID, ttl=2)
        for r_newc in r_newcs:
            r_tman = KWalker(r_newc)
            await r_tman.join(SAMPLING_SERVICE_ID)
            r_kwalkers.append(r_tman)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return kwalker, r_kwalkers

    try:
        yield _init_kwalker
    finally:
        await kwalker.leave()
        for r_kwalker in r_kwalkers:
            await r_kwalker.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_extreme_membership, init_kwalker, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    kwalker, r_kwalkers = await init_kwalker(newc, r_newcs)


@pytest.mark.asyncio
async def test_publish_unpublish(init_extreme_membership, init_kwalker):
    newc, r_newcs = await init_extreme_membership(1)
    kwalker, r_kwalkers = await init_kwalker(newc, r_newcs)

    data = b"test-data"
    data_id = get_random_id()
    await kwalker.publish(data_id, data)

    with pytest.raises(KeyError):
        await kwalker.publish(data_id, data)

    with pytest.raises(KeyError):
        await kwalker.publish(data_id, data + b"more-data")

    with pytest.raises(KeyError):
        await kwalker.unpublish(data_id + "wrong-id")

    await kwalker.unpublish(data_id)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "amount,replication_percent",
    [
        (amount, repl_percent)
        for amount in [
            GossipConfig.LOCAL_VIEW_SIZE + 1,
            GossipConfig.LOCAL_VIEW_SIZE + 5,
            GossipConfig.LOCAL_VIEW_SIZE + 100,
        ]
        for repl_percent in [0.3, 0.6, 0.9]
    ],
)
async def test_search(
    init_extreme_membership, init_kwalker, amount, replication_percent
):
    newc, r_newcs = await init_extreme_membership(amount)
    kwalker, r_kwalkers = await init_kwalker(newc, r_newcs)
    data_id = get_random_id()
    data = b"data"
    replication_amount = ceil(amount * replication_percent)

    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 10)
    for r_kwalker in random.sample(r_kwalkers, replication_amount):
        await r_kwalker.publish(data_id, data)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 5)

    found_data = None
    random_walks_amount = 0
    while not found_data and random_walks_amount <= (amount - replication_amount) + 1:
        try:
            found_data = await kwalker.search(data_id)
        finally:
            random_walks_amount += 1
    assert found_data


@pytest.mark.asyncio
async def test_search_fail(init_extreme_membership, init_kwalker):
    newc, r_newcs = await init_extreme_membership(100)
    kwalker, r_kwalkers = await init_kwalker(newc, r_newcs)
    data_id = get_random_id()
    result = await kwalker.search(data_id)
    assert not result



import asyncio
from typing import Any, Dict, List, Tuple

from rpcudp.protocol import RPCProtocol

from unsserv.common.rpc.structs import Message
from unsserv.common.structs import Node
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.typing import Handler
from unsserv.common.utils import parse_message


class RPCRegister:
    rpc_register: Dict = {}

    @staticmethod
    def get_rpc(node):
        rpc = RPCRegister.rpc_register.get(node, RPC(node))
        RPCRegister.rpc_register[node] = rpc
        return rpc


class RPC(RPCProtocol):
    my_node: Node
    registered_services: Dict[Node, Handler]

    def __init__(self, node: Node):
        RPCProtocol.__init__(self, GossipConfig.RPC_TIMEOUT)
        self.my_node = node
        self.registered_services = {}

    async def call_send_message(self, destination: Node, message: Message) -> Any:
        rpc_result = await self.send_message(destination.address_info, message)
        return self._handle_call_response(rpc_result)

    async def rpc_send_message(self, node: Node, raw_message: List) -> Any:
        message = parse_message(raw_message)
        return await self.registered_services[message.service_id](message)

    async def register_service(self, service_id: Any, handler: Handler):
        if service_id in self.registered_services:
            raise ValueError("Service ID already registered")
        self.registered_services[service_id] = handler

        if (
            len(self.registered_services) == 1
        ):  # activate when first service is registered
            await self._start()

    async def unregister_service(self, service_id: Any):
        if service_id in self.registered_services:
            del self.registered_services[service_id]
        if (
            len(self.registered_services) == 0
        ):  # deactivate when last service is unregistered
            await self._stop()

    async def _start(self):
        (
            self._transport,
            protocol,
        ) = await asyncio.get_event_loop().create_datagram_endpoint(
            lambda: self, self.my_node.address_info
        )

    async def _stop(self):
        if self._transport:
            self._transport.close()
            self._transport = None

    def _handle_call_response(self, result: Tuple[int, Any]) -> Any:
        """
        If we get a response, returns it.

        Otherwise raise error.
        """
        if not result[0]:
            raise ConnectionError(
                "RPC protocol error. Connection failed or invalid value received"
            )
        return result[1]



from typing import Dict, Any

from unsserv.common.utils import IConfig


class DoubleLayeredConfig(IConfig):
    TTL = 10
    ACTIVE_VIEW_SIZE = 6
    MAINTENANCE_SLEEP = 1

    def load_from_dict(self, config_dict: Dict[str, Any]):
        self.TTL = config_dict.get("ttl", DoubleLayeredConfig.TTL)
        self.ACTIVE_VIEW_SIZE = config_dict.get(
            "active_view_size", DoubleLayeredConfig.ACTIVE_VIEW_SIZE
        )
        self.MAINTENANCE_SLEEP = config_dict.get(
            "MAINTENANCE_SLEEP", DoubleLayeredConfig.MAINTENANCE_SLEEP
        )



import asyncio
from abc import ABC, abstractmethod
from dataclasses import is_dataclass, asdict
from enum import IntEnum
from typing import Any, Tuple, Sequence, Dict, Callable

from unsserv.common.rpc.rpc import RPCRegister, RPC
from unsserv.common.rpc.structs import Message
from unsserv.common.structs import Node

Command = IntEnum
Data = Any
Handler = Callable[..., Any]


class ITranscoder(ABC):
    my_node: Node
    service_id: str

    def __init__(self, my_node: Node, service_id: str):
        self.my_node = my_node
        self.service_id = service_id

    @abstractmethod
    def encode(self, command: Command, *data: Data) -> Message:
        pass

    @abstractmethod
    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        pass


class AProtocol:
    my_node: Node
    service_id: str
    _rpc: RPC
    _transcoder: ITranscoder
    _handlers: Dict[Command, Handler]

    _running: bool

    def __init__(self, my_node: Node):
        self.my_node = my_node
        self._rpc = RPCRegister.get_rpc(my_node)
        self._handlers = {}

        self._running = False

    async def start(self, service_id: str):
        if self._running:
            raise RuntimeError("Protocol already running")
        self.service_id = service_id
        self._transcoder = self._get_new_transcoder()
        await self._rpc.register_service(service_id, self.handle_rpc)
        self._running = True

    async def stop(self):
        if self._running:
            await self._rpc.unregister_service(self.service_id)
            self._running = False

    async def handle_rpc(self, message: Message):
        command, data = self._transcoder.decode(message)
        handler = self._handlers[command]
        if asyncio.iscoroutinefunction(handler):
            response = await handler(message.node, *data)
            return self._encode_response(response)
        else:
            response = handler(message.node, *data)
            return self._encode_response(response)

    def _encode_response(self, response: Any) -> Any:
        if isinstance(response, list):
            return [self._encode_response(response_item) for response_item in response]
        elif isinstance(response, tuple):
            return (self._encode_response(response_item) for response_item in response)
        elif hasattr(response, "encode"):
            return response.encode()
        elif is_dataclass(response):
            return asdict(response)
        elif isinstance(response, set):
            return list(response)
        return response

    @abstractmethod
    def _get_new_transcoder(self):
        """
        Method for initializing ITranscoder, because every Protocol implements
        its own.

        :return:
        """
        pass



import asyncio
from functools import partial
from typing import Any, Dict

import pytest

from tests.utils import init_extreme_membership
from unsserv.common.gossip.config import GossipConfig
from unsserv.common.structs import Node
from unsserv.extreme.dissemination.lpbcast.lpbcast import Lpbcast

init_extreme_membership = init_extreme_membership  # for flake8 compliance
DISSEMINATION_SERVICE_ID = "lpbcast"

lpbcast_events: Dict[Node, asyncio.Event] = {}


async def dissemination_handler(node: Node, data: Any):
    lpbcast_events[node].set()


@pytest.mark.asyncio
@pytest.fixture
async def init_lpbcast():
    lpbcast = None
    r_lpbcasts = []

    async def _init_lpbcast(newc, r_newcs):
        nonlocal lpbcast, r_lpbcasts
        lpbcast = Lpbcast(newc)
        lpbcast_events[newc.my_node] = asyncio.Event()
        await lpbcast.join(
            DISSEMINATION_SERVICE_ID,
            broadcast_handler=partial(dissemination_handler, lpbcast.my_node),
        )
        for r_newc in r_newcs:
            r_lpbcast = Lpbcast(r_newc)
            lpbcast_events[r_newc.my_node] = asyncio.Event()
            await r_lpbcast.join(
                DISSEMINATION_SERVICE_ID,
                broadcast_handler=partial(dissemination_handler, r_lpbcast.my_node),
            )
            r_lpbcasts.append(r_lpbcast)
        await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 7)
        return lpbcast, r_lpbcasts

    try:
        yield _init_lpbcast
    finally:
        await lpbcast.leave()
        for r_lpbcast in r_lpbcasts:
            await r_lpbcast.leave()


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_start_stop(init_extreme_membership, init_lpbcast, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    lpbcast, r_lpbcasts = await init_lpbcast(newc, r_newcs)


@pytest.mark.asyncio
@pytest.mark.parametrize("amount", [1, 5, 100])
async def test_broadcast(init_extreme_membership, init_lpbcast, amount):
    newc, r_newcs = await init_extreme_membership(amount)
    lpbcast, r_lpbcasts = await init_lpbcast(newc, r_newcs)

    data = b"data"
    await lpbcast.broadcast(data)
    await asyncio.sleep(GossipConfig.GOSSIPING_FREQUENCY * 15)

    lpbcast_events_received = [
        lpbcast_events[r_lpbcast.my_node].is_set() for r_lpbcast in r_lpbcasts
    ]
    assert int(amount * 0.75) <= sum(lpbcast_events_received)



from unsserv.common.structs import Node
from unsserv.extreme.dissemination.lpbcast.typing import Digest
from dataclasses import dataclass


@dataclass
class Event:
    id: str
    data: bytes
    origin: Node
    digest: Digest
